import pickle
import sys

import numpy as np

import foscat.backend as bk
import foscat.FoCUS as FOC

# Vérifier si TensorFlow est importé et défini
tf_defined = "tensorflow" in sys.modules

if tf_defined:
    import tensorflow as tf

    tf_function = (
        tf.function
    )  # Facultatif : si vous voulez utiliser TensorFlow dans ce script
else:

    def tf_function(func):
        return func


def read(filename):
    thescat = scat1D(1, 1, 1, 1, 1, [0], [0])
    return thescat.read(filename)


class scat1D:
    def __init__(self, p00, s0, s1, s2, s2l, j1, j2, cross=False, backend=None):
        self.P00 = p00
        self.S0 = s0
        self.S1 = s1
        self.S2 = s2
        self.S2L = s2l
        self.j1 = j1
        self.j2 = j2
        self.cross = cross
        self.backend = backend

    # ---------------------------------------------−---------
    def build_flat(self, table):
        shape = table.shape
        ndata = 1
        for k in range(1, len(shape)):
            ndata = ndata * shape[k]
        return self.backend.bk_reshape(table, [shape[0], ndata])

    # ---------------------------------------------−---------
    def flatten(self, S2L=True, P00=True):
        tmp = [self.build_flat(self.S0), self.build_flat(self.S1)]

        if P00:
            tmp = tmp + [self.build_flat(self.P00)]

        tmp = tmp + [self.build_flat(self.S2)]

        if S2L:
            tmp = tmp + [self.build_flat(self.S2L)]

        if isinstance(self.P00, np.ndarray):
            return np.concatenate(tmp, 1)
        else:
            return self.backend.bk_concat(tmp, 1)

    # ---------------------------------------------−---------
    def flatten_name(self, S2L=True, P00=True):

        tmp = ["S0"]

        tmp = tmp + ["S1_%d" % (k) for k in range(self.S1.shape[-1])]

        if P00:
            tmp = tmp + ["P00_%d" % (k) for k in range(self.P00.shape[-1])]

        j1, j2 = self.get_j_idx()

        tmp = tmp + ["S2_%d-%d" % (j1[k], j2[k]) for k in range(self.S2.shape[-1])]

        if S2L:
            tmp = tmp + ["S2L_%d-%d" % (j1[k], j2[k]) for k in range(self.S2.shape[-1])]

        return tmp

    def get_j_idx(self):
        return self.j1, self.j2

    def get_S0(self):
        return self.S0

    def get_S1(self):
        return self.S1

    def get_S2(self):
        return self.S2

    def get_S2L(self):
        return self.S2L

    def get_P00(self):
        return self.P00

    def reset_P00(self):
        self.P00 = 0 * self.P00

    def constant(self):
        return scat1D(
            self.backend.constant(self.P00),
            self.backend.constant(self.S0),
            self.backend.constant(self.S1),
            self.backend.constant(self.S2),
            self.backend.constant(self.S2L),
            self.j1,
            self.j2,
            backend=self.backend,
        )

    def domult(self, x, y):
        if x.dtype == y.dtype:
            return x * y

        if self.backend.bk_is_complex(x):

            return self.backend.bk_complex(
                self.backend.bk_real(x) * y, self.backend.bk_imag(x) * y
            )
        else:
            return self.backend.bk_complex(
                self.backend.bk_real(y) * x, self.backend.bk_imag(y) * x
            )

    def dodiv(self, x, y):
        if x.dtype == y.dtype:
            return x / y
        if self.backend.bk_is_complex(x):

            return self.backend.bk_complex(
                self.backend.bk_real(x) / y, self.backend.bk_imag(x) / y
            )
        else:
            return self.backend.bk_complex(
                x / self.backend.bk_real(y), x / self.backend.bk_imag(y)
            )

    def domin(self, x, y):
        if x.dtype == y.dtype:
            return x - y

        if self.backend.bk_is_complex(x):

            return self.backend.bk_complex(
                self.backend.bk_real(x) - y, self.backend.bk_imag(x) - y
            )
        else:
            return self.backend.bk_complex(
                x - self.backend.bk_real(y), x - self.backend.bk_imag(y)
            )

    def doadd(self, x, y):
        if x.dtype == y.dtype:
            return x + y

        if self.backend.bk_is_complex(x):

            return self.backend.bk_complex(
                self.backend.bk_real(x) + y, self.backend.bk_imag(x) + y
            )
        else:
            return self.backend.bk_complex(
                x + self.backend.bk_real(y), x + self.backend.bk_imag(y)
            )

    def __add__(self, other):
        assert (
            isinstance(other, float)
            or isinstance(other, np.float32)
            or isinstance(other, int)
            or isinstance(other, bool)
            or isinstance(other, scat1D)
        )

        if isinstance(other, scat1D):
            return scat1D(
                self.doadd(self.P00, other.P00),
                self.doadd(self.S0, other.S0),
                self.doadd(self.S1, other.S1),
                self.doadd(self.S2, other.S2),
                self.doadd(self.S2L, other.S2L),
                self.j1,
                self.j2,
                backend=self.backend,
            )
        else:
            return scat1D(
                (self.P00 + other),
                (self.S0 + other),
                (self.S1 + other),
                (self.S2 + other),
                (self.S2L + other),
                self.j1,
                self.j2,
                backend=self.backend,
            )

    def toreal(self, value):
        if value is None:
            return None

        return self.backend.bk_real(value)

    def addcomplex(self, value, amp):
        if value is None:
            return None

        return self.backend.bk_complex(value, amp * value)

    def add_complex(self, amp):
        return scat1D(
            self.addcomplex(self.P00, amp),
            self.addcomplex(self.S0, amp),
            self.addcomplex(self.S1, amp),
            self.addcomplex(self.S2, amp),
            self.addcomplex(self.S2L, amp),
            self.j1,
            self.j2,
            backend=self.backend,
        )

    def real(self):
        return scat1D(
            self.toreal(self.P00),
            self.toreal(self.S0),
            self.toreal(self.S1),
            self.toreal(self.S2),
            self.toreal(self.S2L),
            self.j1,
            self.j2,
            backend=self.backend,
        )

    def __radd__(self, other):
        return self.__add__(other)

    def __truediv__(self, other):
        assert (
            isinstance(other, float)
            or isinstance(other, np.float32)
            or isinstance(other, int)
            or isinstance(other, bool)
            or isinstance(other, scat1D)
        )

        if isinstance(other, scat1D):
            return scat1D(
                self.dodiv(self.P00, other.P00),
                self.dodiv(self.S0, other.S0),
                self.dodiv(self.S1, other.S1),
                self.dodiv(self.S2, other.S2),
                self.dodiv(self.S2L, other.S2L),
                self.j1,
                self.j2,
                backend=self.backend,
            )
        else:
            return scat1D(
                (self.P00 / other),
                (self.S0 / other),
                (self.S1 / other),
                (self.S2 / other),
                (self.S2L / other),
                self.j1,
                self.j2,
                backend=self.backend,
            )

    def __rtruediv__(self, other):
        assert (
            isinstance(other, float)
            or isinstance(other, np.float32)
            or isinstance(other, int)
            or isinstance(other, bool)
            or isinstance(other, scat1D)
        )

        if isinstance(other, scat1D):
            return scat1D(
                self.dodiv(other.P00, self.P00),
                self.dodiv(other.S0, self.S0),
                self.dodiv(other.S1, self.S1),
                self.dodiv(other.S2, self.S2),
                self.dodiv(other.S2L, self.S2L),
                self.j1,
                self.j2,
                backend=self.backend,
            )
        else:
            return scat1D(
                (other / self.P00),
                (other / self.S0),
                (other / self.S1),
                (other / self.S2),
                (other / self.S2L),
                self.j1,
                self.j2,
                backend=self.backend,
            )

    def __sub__(self, other):
        assert (
            isinstance(other, float)
            or isinstance(other, np.float32)
            or isinstance(other, int)
            or isinstance(other, bool)
            or isinstance(other, scat1D)
        )

        if isinstance(other, scat1D):
            return scat1D(
                self.domin(self.P00, other.P00),
                self.domin(self.S0, other.S0),
                self.domin(self.S1, other.S1),
                self.domin(self.S2, other.S2),
                self.domin(self.S2L, other.S2L),
                self.j1,
                self.j2,
                backend=self.backend,
            )
        else:
            return scat1D(
                (self.P00 - other),
                (self.S0 - other),
                (self.S1 - other),
                (self.S2 - other),
                (self.S2L - other),
                self.j1,
                self.j2,
                backend=self.backend,
            )

    def __rsub__(self, other):
        assert (
            isinstance(other, float)
            or isinstance(other, np.float32)
            or isinstance(other, int)
            or isinstance(other, bool)
            or isinstance(other, scat1D)
        )

        if isinstance(other, scat1D):
            return scat1D(
                self.domin(other.P00, self.P00),
                self.domin(other.S0, self.S0),
                self.domin(other.S1, self.S1),
                self.domin(other.S2, self.S2),
                self.domin(other.S2L, self.S2L),
                self.j1,
                self.j2,
                backend=self.backend,
            )
        else:
            return scat1D(
                (other - self.P00),
                (other - self.S0),
                (other - self.S1),
                (other - self.S2),
                (other - self.S2L),
                self.j1,
                self.j2,
                backend=self.backend,
            )

    def __mul__(self, other):
        assert (
            isinstance(other, float)
            or isinstance(other, np.float32)
            or isinstance(other, int)
            or isinstance(other, bool)
            or isinstance(other, scat1D)
        )

        if isinstance(other, scat1D):
            return scat1D(
                self.domult(self.P00, other.P00),
                self.domult(self.S0, other.S0),
                self.domult(self.S1, other.S1),
                self.domult(self.S2, other.S2),
                self.domult(self.S2L, other.S2L),
                self.j1,
                self.j2,
                backend=self.backend,
            )
        else:
            return scat1D(
                (self.P00 * other),
                (self.S0 * other),
                (self.S1 * other),
                (self.S2 * other),
                (self.S2L * other),
                self.j1,
                self.j2,
                backend=self.backend,
            )

    def relu(self):
        return scat1D(
            self.backend.bk_relu(self.P00),
            self.backend.bk_relu(self.S0),
            self.backend.bk_relu(self.S1),
            self.backend.bk_relu(self.S2),
            self.backend.bk_relu(self.S2L),
            self.j1,
            self.j2,
            backend=self.backend,
        )

    def __rmul__(self, other):
        assert (
            isinstance(other, float)
            or isinstance(other, np.float32)
            or isinstance(other, int)
            or isinstance(other, bool)
            or isinstance(other, scat1D)
        )

        if isinstance(other, scat1D):
            return scat1D(
                self.domult(self.P00, other.P00),
                self.domult(self.S0, other.S0),
                self.domult(self.S1, other.S1),
                self.domult(self.S2, other.S2),
                self.domult(self.S2L, other.S2L),
                self.j1,
                self.j2,
                backend=self.backend,
            )
        else:
            return scat1D(
                (self.P00 * other),
                (self.S0 * other),
                (self.S1 * other),
                (self.S2 * other),
                (self.S2L * other),
                self.j1,
                self.j2,
                backend=self.backend,
            )

    def l1_abs(self, x):
        y = self.get_np(x)
        if self.backend.bk_is_complex(y):
            tmp = y.real * y.real + y.imag * y.imag
            tmp = np.sign(tmp) * np.sqrt(np.fabs(tmp))
            y = tmp

        return y

    def plot(self, name=None, hold=True, color="blue", lw=1, legend=True):

        import matplotlib.pyplot as plt

        j1, j2 = self.get_j_idx()

        if name is None:
            name = ""

        if hold:
            plt.figure(figsize=(16, 8))

        test = None
        plt.subplot(2, 2, 1)
        tmp = abs(self.get_np(self.S1))
        if len(tmp.shape) == 4:
            for i1 in range(tmp.shape[0]):
                for i2 in range(tmp.shape[1]):
                    if test is None:
                        test = 1
                        plt.plot(
                            tmp[i1, i2, :],
                            color=color,
                            label=r"%s $S_{1}$" % (name),
                            lw=lw,
                        )
                    else:
                        plt.plot(tmp[i1, i2, :], color=color, lw=lw)
        else:
            for i1 in range(tmp.shape[0]):
                if test is None:
                    test = 1
                    plt.plot(
                        tmp[i1, :], color=color, label=r"%s $S_{1}$" % (name), lw=lw
                    )
                else:
                    plt.plot(tmp[i1, :], color=color, lw=lw)
        plt.yscale("log")
        plt.ylabel("S1")
        plt.xlabel(r"$j_{1}$")
        plt.legend()

        test = None
        plt.subplot(2, 2, 2)
        tmp = abs(self.get_np(self.P00))
        if len(tmp.shape) == 4:
            for i1 in range(tmp.shape[0]):
                for i2 in range(tmp.shape[0]):
                    if test is None:
                        test = 1
                        plt.plot(
                            tmp[i1, i2, :],
                            color=color,
                            label=r"%s $P_{00}$" % (name),
                            lw=lw,
                        )
                    else:
                        plt.plot(tmp[i1, i2, :], color=color, lw=lw)
        else:
            for i1 in range(tmp.shape[0]):
                if test is None:
                    test = 1
                    plt.plot(
                        tmp[i1, :], color=color, label=r"%s $P_{00}$" % (name), lw=lw
                    )
                else:
                    plt.plot(tmp[i1, :], color=color, lw=lw)
        plt.yscale("log")
        plt.ylabel("P00")
        plt.xlabel(r"$j_{1}$")
        plt.legend()

        ax1 = plt.subplot(2, 2, 3)
        ax2 = ax1.twiny()
        n = 0
        tmp = abs(self.get_np(self.S2))
        lname = r"%s $S_{2}$" % (name)
        ax1.set_ylabel(r"$S_{2}$ [L1 norm]")
        test = None
        tabx = []
        tabnx = []
        tab2x = []
        tab2nx = []
        if len(tmp.shape) == 5:
            for i0 in range(tmp.shape[0]):
                for i1 in range(tmp.shape[1]):
                    for i2 in range(j1.max() + 1):
                        if j2[j1 == i2].shape[0] == 1:
                            ax1.plot(
                                j2[j1 == i2] + n,
                                tmp[i0, i1, j1 == i2],
                                ".",
                                color=color,
                                lw=lw,
                            )
                        else:
                            if legend and test is None:
                                ax1.plot(
                                    j2[j1 == i2] + n,
                                    tmp[i0, i1, j1 == i2],
                                    color=color,
                                    label=lname,
                                    lw=lw,
                                )
                                test = 1
                            ax1.plot(
                                j2[j1 == i2] + n,
                                tmp[i0, i1, j1 == i2],
                                color=color,
                                lw=lw,
                            )
                    tabnx = tabnx + [r"%d" % (k) for k in j2[j1 == i2]]
                    tabx = tabx + [k + n for k in j2[j1 == i2]]
                    tab2x = tab2x + [(j2[j1 == i2] + n).mean()]
                    tab2nx = tab2nx + ["%d" % (i2)]
                    ax1.axvline((j2[j1 == i2] + n).max() + 0.5, ls=":", color="gray")
                    n = n + j2[j1 == i2].shape[0] - 1
        else:
            for i0 in range(tmp.shape[0]):
                for i2 in range(j1.max() + 1):
                    if j2[j1 == i2].shape[0] == 1:
                        ax1.plot(
                            j2[j1 == i2] + n, tmp[i0, j1 == i2], ".", color=color, lw=lw
                        )
                    else:
                        if legend and test is None:
                            ax1.plot(
                                j2[j1 == i2] + n,
                                tmp[i0, j1 == i2],
                                color=color,
                                label=lname,
                                lw=lw,
                            )
                            test = 1
                        ax1.plot(
                            j2[j1 == i2] + n, tmp[i0, j1 == i2], color=color, lw=lw
                        )
                    tabnx = tabnx + [r"%d" % (k) for k in j2[j1 == i2]]
                    tabx = tabx + [k + n for k in j2[j1 == i2]]
                    tab2x = tab2x + [(j2[j1 == i2] + n).mean()]
                    tab2nx = tab2nx + ["%d" % (i2)]
                    ax1.axvline((j2[j1 == i2] + n).max() + 0.5, ls=":", color="gray")
                    n = n + j2[j1 == i2].shape[0] - 1
        plt.yscale("log")
        ax1.set_xlim(0, n + 2)
        ax1.set_xticks(tabx)
        ax1.set_xticklabels(tabnx, fontsize=6)
        ax1.set_xlabel(r"$j_{2}$ ", fontsize=6)

        # Move twinned axis ticks and label from top to bottom
        ax2.xaxis.set_ticks_position("bottom")
        ax2.xaxis.set_label_position("bottom")

        # Offset the twin axis below the host
        ax2.spines["bottom"].set_position(("axes", -0.15))

        # Turn on the frame for the twin axis, but then hide all
        # but the bottom spine
        ax2.set_frame_on(True)
        ax2.patch.set_visible(False)

        for sp in ax2.spines.values():
            sp.set_visible(False)
        ax2.spines["bottom"].set_visible(True)
        ax2.set_xlim(0, n + 2)
        ax2.set_xticks(tab2x)
        ax2.set_xticklabels(tab2nx, fontsize=6)
        ax2.set_xlabel(r"$j_{1}$", fontsize=6)
        ax1.legend(frameon=0)

        ax1 = plt.subplot(2, 2, 4)
        ax2 = ax1.twiny()
        n = 0
        tmp = abs(self.get_np(self.S2L))
        lname = r"%s $S2_{2}$" % (name)
        ax1.set_ylabel(r"$S_{2}$ [L2 norm]")
        test = None
        tabx = []
        tabnx = []
        tab2x = []
        tab2nx = []
        if len(tmp.shape) == 5:
            for i0 in range(tmp.shape[0]):
                for i1 in range(tmp.shape[1]):
                    for i2 in range(j1.max() + 1):
                        if j2[j1 == i2].shape[0] == 1:
                            ax1.plot(
                                j2[j1 == i2] + n,
                                tmp[i0, i1, j1 == i2],
                                ".",
                                color=color,
                                lw=lw,
                            )
                        else:
                            if legend and test is None:
                                ax1.plot(
                                    j2[j1 == i2] + n,
                                    tmp[i0, i1, j1 == i2],
                                    color=color,
                                    label=lname,
                                    lw=lw,
                                )
                                test = 1
                            ax1.plot(
                                j2[j1 == i2] + n,
                                tmp[i0, i1, j1 == i2],
                                color=color,
                                lw=lw,
                            )
                        tabnx = tabnx + [r"%d" % (k) for k in j2[j1 == i2]]
                        tabx = tabx + [k + n for k in j2[j1 == i2]]
                        tab2x = tab2x + [(j2[j1 == i2] + n).mean()]
                        tab2nx = tab2nx + ["%d" % (i2)]
                        ax1.axvline(
                            (j2[j1 == i2] + n).max() + 0.5, ls=":", color="gray"
                        )
                        n = n + j2[j1 == i2].shape[0] - 1
        else:
            for i0 in range(tmp.shape[0]):
                for i2 in range(j1.max() + 1):
                    if j2[j1 == i2].shape[0] == 1:
                        ax1.plot(
                            j2[j1 == i2] + n, tmp[i0, j1 == i2], ".", color=color, lw=lw
                        )
                    else:
                        if legend and test is None:
                            ax1.plot(
                                j2[j1 == i2] + n,
                                tmp[i0, j1 == i2],
                                color=color,
                                label=lname,
                                lw=lw,
                            )
                            test = 1
                        ax1.plot(
                            j2[j1 == i2] + n, tmp[i0, j1 == i2], color=color, lw=lw
                        )
                    tabnx = tabnx + [r"%d" % (k) for k in j2[j1 == i2]]
                    tabx = tabx + [k + n for k in j2[j1 == i2]]
                    tab2x = tab2x + [(j2[j1 == i2] + n).mean()]
                    tab2nx = tab2nx + ["%d" % (i2)]
                    ax1.axvline((j2[j1 == i2] + n).max() + 0.5, ls=":", color="gray")
                    n = n + j2[j1 == i2].shape[0] - 1
        plt.yscale("log")
        ax1.set_xlim(-1, n + 3)
        ax1.set_xticks(tabx)
        ax1.set_xticklabels(tabnx, fontsize=6)
        ax1.set_xlabel(r"$j_{2}$", fontsize=6)

        # Move twinned axis ticks and label from top to bottom
        ax2.xaxis.set_ticks_position("bottom")
        ax2.xaxis.set_label_position("bottom")

        # Offset the twin axis below the host
        ax2.spines["bottom"].set_position(("axes", -0.15))

        # Turn on the frame for the twin axis, but then hide all
        # but the bottom spine
        ax2.set_frame_on(True)
        ax2.patch.set_visible(False)

        for sp in ax2.spines.values():
            sp.set_visible(False)
        ax2.spines["bottom"].set_visible(True)
        ax2.set_xlim(0, n + 3)
        ax2.set_xticks(tab2x)
        ax2.set_xticklabels(tab2nx, fontsize=6)
        ax2.set_xlabel(r"$j_{1}$", fontsize=6)
        ax1.legend(frameon=0)

    def save(self, filename):
        outlist = [
            self.get_S0().numpy(),
            self.get_S1().numpy(),
            self.get_S2().numpy(),
            self.get_S2L().numpy(),
            self.get_P00().numpy(),
            self.j1,
            self.j2,
        ]

        myout = open("%s.pkl" % (filename), "wb")
        pickle.dump(outlist, myout)
        myout.close()

    def read(self, filename):

        outlist = pickle.load(open("%s.pkl" % (filename), "rb"))

        return scat1D(
            outlist[4],
            outlist[0],
            outlist[1],
            outlist[2],
            outlist[3],
            outlist[5],
            outlist[6],
            backend=bk.foscat_backend("numpy"),
        )

    def get_np(self, x):
        if isinstance(x, np.ndarray):
            return x
        else:
            return x.numpy()

    def std(self):
        return np.sqrt(
            (
                (abs(self.get_np(self.S0)).std()) ** 2
                + (abs(self.get_np(self.S1)).std()) ** 2
                + (abs(self.get_np(self.S2)).std()) ** 2
                + (abs(self.get_np(self.S2L)).std()) ** 2
                + (abs(self.get_np(self.P00)).std()) ** 2
            )
            / 4
        )

    def mean(self):
        return (
            abs(
                self.get_np(self.S0).mean()
                + self.get_np(self.S1).mean()
                + self.get_np(self.S2).mean()
                + self.get_np(self.S2L).mean()
                + self.get_np(self.P00).mean()
            )
            / 3
        )

    # ---------------------------------------------−---------
    def cleanval(self, x):
        x = x.numpy()
        x[~np.isfinite(x)] = np.median(x[np.isfinite(x)])
        return x

    def filter_inf(self):
        S1 = self.cleanval(self.S1)
        S0 = self.cleanval(self.S0)
        P00 = self.cleanval(self.P00)
        S2 = self.cleanval(self.S2)
        S2L = self.cleanval(self.S2L)

        return scat1D(P00, S0, S1, S2, S2L, self.j1, self.j2, backend=self.backend)

    # ---------------------------------------------−---------
    def interp(self, nscale, extend=False, constant=False, threshold=1e30):

        if nscale + 2 > self.S1.shape[1]:
            print(
                "Can not *interp* %d with a statistic described over %d"
                % (nscale, self.S1.shape[1])
            )
            return scat1D(
                self.P00,
                self.S0,
                self.S1,
                self.S2,
                self.S2L,
                self.j1,
                self.j2,
                backend=self.backend,
            )
        if isinstance(self.S1, np.ndarray):
            s1 = self.S1
            p0 = self.P00
            s2 = self.S2
            s2l = self.S2L
        else:
            s1 = self.S1.numpy()
            p0 = self.P00.numpy()
            s2 = self.S2.numpy()
            s2l = self.S2L.numpy()

        if isinstance(threshold, scat1D):
            if isinstance(threshold.S1, np.ndarray):
                s1th = threshold.S1
                p0th = threshold.P00
                s2th = threshold.S2
                s2lth = threshold.S2L
            else:
                s1th = threshold.S1.numpy()
                p0th = threshold.P00.numpy()
                s2th = threshold.S2.numpy()
                s2lth = threshold.S2L.numpy()
        else:
            s1th = threshold + 0 * s1
            p0th = threshold + 0 * p0
            s2th = threshold + 0 * s2
            s2lth = threshold + 0 * s2l

        for k in range(nscale):
            if constant:
                s1[:, nscale - 1 - k, :] = s1[:, nscale - k, :]
                p0[:, nscale - 1 - k, :] = p0[:, nscale - k, :]
            else:
                idx = np.where(
                    (s1[:, nscale + 1 - k, :] > 0)
                    * (s1[:, nscale + 2 - k, :] > 0)
                    * (s1[:, nscale - k, :] < s1th[:, nscale - k, :])
                )
                if len(idx[0]) > 0:
                    s1[idx[0], nscale - 1 - k, idx[1]] = np.exp(
                        3 * np.log(s1[idx[0], nscale + 1 - k, idx[1]])
                        - 2 * np.log(s1[idx[0], nscale + 2 - k, idx[1]])
                    )
                idx = np.where(
                    (s1[:, nscale - k, :] > 0)
                    * (s1[:, nscale + 1 - k, :] > 0)
                    * (s1[:, nscale - 1 - k, :] < s1th[:, nscale - 1 - k, :])
                )
                if len(idx[0]) > 0:
                    s1[idx[0], nscale - 1 - k, idx[1]] = np.exp(
                        2 * np.log(s1[idx[0], nscale - k, idx[1]])
                        - np.log(s1[idx[0], nscale + 1 - k, idx[1]])
                    )

                idx = np.where(
                    (p0[:, nscale + 1 - k, :] > 0)
                    * (p0[:, nscale + 2 - k, :] > 0)
                    * (p0[:, nscale - k, :] < p0th[:, nscale - k, :])
                )
                if len(idx[0]) > 0:
                    p0[idx[0], nscale - 1 - k, idx[1]] = np.exp(
                        3 * np.log(p0[idx[0], nscale + 1 - k, idx[1]])
                        - 2 * np.log(p0[idx[0], nscale + 2 - k, idx[1]])
                    )

                idx = np.where(
                    (p0[:, nscale - k, :] > 0)
                    * (p0[:, nscale + 1 - k, :] > 0)
                    * (p0[:, nscale - 1 - k, :] < p0th[:, nscale - 1 - k, :])
                )
                if len(idx[0]) > 0:
                    p0[idx[0], nscale - 1 - k, idx[1]] = np.exp(
                        2 * np.log(p0[idx[0], nscale - k, idx[1]])
                        - np.log(p0[idx[0], nscale + 1 - k, idx[1]])
                    )

        j1, j2 = self.get_j_idx()

        for k in range(nscale):

            """
            i0=np.where((j1==nscale-1-k)*(j2>=nscale+1-k))[0]
            i1=np.where((j1==nscale-k)*(j2>=nscale+1-k))[0]
            i2=np.where((j1==nscale+1-k)*(j2>=nscale+1-k))[0]
            print(i0,j1[i0],j2[i0],j1[i1],j2[i1],j1[i2],j2[i2])
            s2[:,i0]=np.exp(2*np.log(s2[:,i1])-np.log(s2[:,i2]))
            s2l[:,i0]=np.exp(2*np.log(s2l[:,i1])-np.log(s2l[:,i2]))
            """

            for l_orient in range(nscale - k):
                i0 = np.where(
                    (j1 == nscale - 1 - k - l_orient) * (j2 == nscale - 1 - k)
                )[0]
                i1 = np.where((j1 == nscale - 1 - k - l_orient) * (j2 == nscale - k))[0]
                i2 = np.where(
                    (j1 == nscale - 1 - k - l_orient) * (j2 == nscale + 1 - k)
                )[0]
                i3 = np.where(
                    (j1 == nscale - 1 - k - l_orient) * (j2 == nscale + 2 - k)
                )[0]

                if constant:
                    s2[:, i0] = s2[:, i1]
                    s2l[:, i0] = s2l[:, i1]
                else:
                    idx = np.where(
                        (s2[:, i2] > 0) * (s2[:, i3] > 0) * (s2[:, i2] < s2th[:, i2])
                    )
                    if len(idx[0]) > 0:
                        s2[idx[0], i0, idx[1], idx[2]] = np.exp(
                            3 * np.log(s2[idx[0], i2, idx[1], idx[2]])
                            - 2 * np.log(s2[idx[0], i3, idx[1], idx[2]])
                        )

                    idx = np.where(
                        (s2[:, i1] > 0) * (s2[:, i2] > 0) * (s2[:, i1] < s2th[:, i1])
                    )
                    if len(idx[0]) > 0:
                        s2[idx[0], i0, idx[1], idx[2]] = np.exp(
                            2 * np.log(s2[idx[0], i1, idx[1], idx[2]])
                            - np.log(s2[idx[0], i2, idx[1], idx[2]])
                        )

                    idx = np.where(
                        (s2l[:, i2] > 0)
                        * (s2l[:, i3] > 0)
                        * (s2l[:, i2] < s2lth[:, i2])
                    )
                    if len(idx[0]) > 0:
                        s2l[idx[0], i0, idx[1], idx[2]] = np.exp(
                            3 * np.log(s2l[idx[0], i2, idx[1], idx[2]])
                            - 2 * np.log(s2l[idx[0], i3, idx[1], idx[2]])
                        )

                    idx = np.where(
                        (s2l[:, i1] > 0)
                        * (s2l[:, i2] > 0)
                        * (s2l[:, i1] < s2lth[:, i1])
                    )
                    if len(idx[0]) > 0:
                        s2l[idx[0], i0, idx[1], idx[2]] = np.exp(
                            2 * np.log(s2l[idx[0], i1, idx[1], idx[2]])
                            - np.log(s2l[idx[0], i2, idx[1], idx[2]])
                        )

        if extend:
            for k in range(nscale):
                for l_orient in range(1, nscale):
                    i0 = np.where(
                        (j1 == 2 * nscale - 1 - k)
                        * (j2 == 2 * nscale - 1 - k - l_orient)
                    )[0]
                    i1 = np.where(
                        (j1 == 2 * nscale - 1 - k) * (j2 == 2 * nscale - k - l_orient)
                    )[0]
                    i2 = np.where(
                        (j1 == 2 * nscale - 1 - k)
                        * (j2 == 2 * nscale + 1 - k - l_orient)
                    )[0]
                    i3 = np.where(
                        (j1 == 2 * nscale - 1 - k)
                        * (j2 == 2 * nscale + 2 - k - l_orient)
                    )[0]
                    if constant:
                        s2[:, i0] = s2[:, i1]
                        s2l[:, i0] = s2l[:, i1]
                    else:
                        idx = np.where(
                            (s2[:, i2] > 0)
                            * (s2[:, i3] > 0)
                            * (s2[:, i2] < s2th[:, i2])
                        )
                        if len(idx[0]) > 0:
                            s2[idx[0], i0, idx[1], idx[2]] = np.exp(
                                3 * np.log(s2[idx[0], i2, idx[1], idx[2]])
                                - 2 * np.log(s2[idx[0], i3, idx[1], idx[2]])
                            )
                        idx = np.where(
                            (s2[:, i1] > 0)
                            * (s2[:, i2] > 0)
                            * (s2[:, i1] < s2th[:, i1])
                        )
                        if len(idx[0]) > 0:
                            s2[idx[0], i0, idx[1], idx[2]] = np.exp(
                                2 * np.log(s2[idx[0], i1, idx[1], idx[2]])
                                - np.log(s2[idx[0], i2, idx[1], idx[2]])
                            )

                        idx = np.where(
                            (s2l[:, i2] > 0)
                            * (s2l[:, i3] > 0)
                            * (s2l[:, i2] < s2lth[:, i2])
                        )
                        if len(idx[0]) > 0:
                            s2l[idx[0], i0, idx[1], idx[2]] = np.exp(
                                3 * np.log(s2l[idx[0], i2, idx[1], idx[2]])
                                - 2 * np.log(s2l[idx[0], i3, idx[1], idx[2]])
                            )
                        idx = np.where(
                            (s2l[:, i1] > 0)
                            * (s2l[:, i2] > 0)
                            * (s2l[:, i1] < s2lth[:, i1])
                        )
                        if len(idx[0]) > 0:
                            s2l[idx[0], i0, idx[1], idx[2]] = np.exp(
                                2 * np.log(s2l[idx[0], i1, idx[1], idx[2]])
                                - np.log(s2l[idx[0], i2, idx[1], idx[2]])
                            )

        s1[np.isnan(s1)] = 0.0
        p0[np.isnan(p0)] = 0.0
        s2[np.isnan(s2)] = 0.0
        s2l[np.isnan(s2l)] = 0.0

        return scat1D(
            self.backend.constant(p0),
            self.S0,
            self.backend.constant(s1),
            self.backend.constant(s2),
            self.backend.constant(s2l),
            self.j1,
            self.j2,
            backend=self.backend,
        )

    # ---------------------------------------------−---------
    def model(self, i__y, add=0, dx=3, dell=2, weigth=None, inverse=False):

        if i__y.shape[0] < dx + 1:
            l__dx = i__y.shape[0] - 1
        else:
            l__dx = dx

        if i__y.shape[0] < dell:
            l__dell = 0
        else:
            l__dell = dell

        if l__dx < 2:
            res = np.zeros([i__y.shape[0] + add])
            if inverse:
                res[:-add] = i__y
            else:
                res[add:] = i__y[0:]
            return res

        if weigth is None:
            w = 2 ** (np.arange(l__dx))
        else:
            if not inverse:
                w = weigth[0:l__dx]
            else:
                w = weigth[-l__dx:]

        x = np.arange(l__dx) + 1
        if not inverse:
            y = np.log(i__y[1 : l__dx + 1])
        else:
            y = np.log(i__y[-(l__dx + 1) : -1])

        r = np.polyfit(x, y, 1, w=w)

        if inverse:
            res = np.exp(r[0] * (np.arange(i__y.shape[0] + add) - 1) + r[1])
            res[: -(l__dell + add)] = i__y[:-l__dell]
        else:
            res = np.exp(r[0] * (np.arange(i__y.shape[0] + add) - add) + r[1])
            res[l__dell + add :] = i__y[l__dell:]
        return res

    def findn(self, n):
        d = np.sqrt(1 + 8 * n)
        return int((d - 1) / 2)

    def findidx(self, s2):
        i1 = np.zeros([s2.shape[1]], dtype="int")
        i2 = np.zeros([s2.shape[1]], dtype="int")
        n = 0
        for k in range(self.findn(s2.shape[1])):
            i1[n : n + k + 1] = np.arange(k + 1)
            i2[n : n + k + 1] = k
            n = n + k + 1
        return i1, i2

    def extrapol_s2(self, add, lnorm=1):
        if lnorm == 1:
            s2 = self.S2.numpy()
        if lnorm == 2:
            s2 = self.S2L.numpy()
        i1, i2 = self.findidx(s2)

        so2 = np.zeros(
            [
                s2.shape[0],
                (self.findn(s2.shape[1]) + add)
                * (self.findn(s2.shape[1]) + add + 1)
                // 2,
                s2.shape[2],
                s2.shape[3],
            ]
        )
        oi1, oi2 = self.findidx(so2)
        for l_orient in range(s2.shape[0]):
            for k in range(self.findn(s2.shape[1])):
                for i in range(s2.shape[2]):
                    for j in range(s2.shape[3]):
                        tmp = self.model(
                            s2[l_orient, i2 == k, i, j],
                            dx=4,
                            dell=1,
                            add=add,
                            weigth=np.array([1, 2, 2, 2]),
                        )
                        tmp[np.isnan(tmp)] = 0.0
                        so2[l_orient, oi2 == k + add, i, j] = tmp

        for l_orient in range(s2.shape[0]):
            for k in range(add + 1, -1, -1):
                lidx = np.where(oi2 - oi1 == k)[0]
                lidx2 = np.where(oi2 - oi1 == k + 1)[0]
                for i in range(s2.shape[2]):
                    for j in range(s2.shape[3]):
                        so2[l_orient, lidx[0 : add + 2 - k], i, j] = so2[
                            l_orient, lidx2[0 : add + 2 - k], i, j
                        ]

        return so2

    def extrapol_s1(self, i_s1, add):
        s1 = i_s1.numpy()
        so1 = np.zeros([s1.shape[0], s1.shape[1] + add, s1.shape[2]])
        for k in range(s1.shape[0]):
            for i in range(s1.shape[2]):
                so1[k, :, i] = self.model(s1[k, :, i], dx=4, dell=1, add=add)
                so1[k, np.isnan(so1[k, :, i]), i] = 0.0
        return so1

    def extrapol(self, add):
        return scat1D(
            self.extrapol_s1(self.P00, add),
            self.S0,
            self.extrapol_s1(self.S1, add),
            self.extrapol_s2(add, lnorm=1),
            self.extrapol_s2(add, lnorm=2),
            self.j1,
            self.j2,
            backend=self.backend,
        )


class funct(FOC.FoCUS):

    def fill(self, im, nullval=0):
        return self.fill_1d(im, nullval=nullval)

    def ud_grade(self, im, nout, axis=0):
        return self.ud_grade_1d(im, nout, axis=axis)

    def up_grade(self, im, nout, axis=0):
        return self.up_grade_1d(im, nout, axis=axis)

    def smooth(self, data, axis=0):
        return self.smooth_1d(data, axis=axis)

    def convol(self, data, axis=0):
        return self.convol_1d(data, axis=axis)

    def eval(
        self,
        image1,
        image2=None,
        mask=None,
        Auto=True,
        s0_off=1e-6,
        Add_R45=False,
        axis=0,
    ):
        # Check input consistency
        if mask is not None:
            if len(image1.shape) == 1:
                if image1.shape[0] != mask.shape[1]:
                    print(
                        "The mask should have the same size than the input timeline to eval Scattering"
                    )
                    return None
            else:
                if image1.shape[1] != mask.shape[1]:
                    print(
                        "The mask should have the same size than the input timeline to eval Scattering"
                    )
                    return None

        ### AUTO OR CROSS
        cross = False
        if image2 is not None:
            cross = True

        # determine jmax and nside corresponding to the input map
        im_shape = image1.shape

        nside = im_shape[len(image1.shape) - 1]
        npix = nside

        jmax = int(np.log(nside) / np.log(2))  # -self.OSTEP

        ### LOCAL VARIABLES (IMAGES and MASK)
        # Check if image1 is [Npix] or [Nbatch,Npix]
        if len(image1.shape) == 1:
            # image1 is [Nbatch, Npix]
            I1 = self.backend.bk_cast(
                self.backend.bk_expand_dims(image1, 0)
            )  # Local image1 [Nbatch, Npix]
            if cross:
                I2 = self.backend.bk_cast(
                    self.backend.bk_expand_dims(image2, 0)
                )  # Local image2 [Nbatch, Npix]
        else:
            I1 = self.backend.bk_cast(image1)
            if cross:
                I2 = self.backend.bk_cast(image2)

        # self.mask is [Nmask, Npix]

        if mask is None:
            vmask = self.backend.bk_ones([1, npix], dtype=self.all_type)
        else:
            vmask = self.backend.bk_cast(mask)  # [Nmask, Npix]

        if self.KERNELSZ > 3:
            # if the kernel size is bigger than 3 increase the binning before smoothing
            l_image1 = self.up_grade_1d(I1, nside * 2, axis=axis + 1)
            vmask = self.up_grade_1d(vmask, nside * 2, axis=1)

            if cross:
                l_image2 = self.up_grade_1d(I2, nside * 2, axis=axis + 1)
        else:
            l_image1 = I1
            if cross:
                l_image2 = I2
        if len(image1.shape) == 1:
            s0 = self.backend.bk_reduce_sum(l_image1 * vmask, axis=axis + 1)
            if cross and not Auto:
                s0 = self.backend.bk_concat(
                    [s0, self.backend.bk_reduce_sum(l_image2 * vmask, axis=axis)]
                )
        else:
            lmask = self.backend.bk_expand_dims(vmask, 0)
            lim = self.backend.bk_expand_dims(l_image1, 1)
            s0 = self.backend.bk_reduce_sum(lim * lmask, axis=axis + 2)
            if cross and not Auto:
                lim = self.backend.bk_expand_dims(l_image2, 1)
                s0 = self.backend.bk_concat(
                    [s0, self.backend.bk_reduce_sum(lim * lmask, axis=axis + 2)]
                )

        s1 = None
        s2 = None
        s2l = None
        p00 = None
        s2j1 = None
        s2j2 = None

        l2_image = None

        for j1 in range(jmax):
            if j1 < jmax - self.OSTEP:  # stop to add scales
                # Convol image along the axis defined by 'axis' using the wavelet defined at
                # the foscat initialisation
                # c_image_real is [....,Npix_j1,....,Norient]
                c_image1 = self.convol_1d(l_image1, axis=axis + 1)
                if cross:
                    c_image2 = self.convol_1d(l_image2, axis=axis + 1)
                else:
                    c_image2 = c_image1

                # Compute (a+ib)*(a+ib)* the last c_image column is the real and imaginary part
                conj = c_image1 * self.backend.bk_conjugate(c_image2)

                if Auto:
                    conj = self.backend.bk_real(conj)

                # Compute l_p00 [....,....,Nmask,1]
                l_p00 = self.backend.bk_expand_dims(
                    self.backend.bk_reduce_sum(conj * vmask, axis=1), -1
                )

                conj = self.backend.bk_L1(conj)

                # Compute l_s1 [....,....,Nmask,1]
                l_s1 = self.backend.bk_expand_dims(
                    self.backend.bk_reduce_sum(conj * vmask, axis=1), -1
                )

                # Concat S1,P00 [....,....,Nmask,j1]
                if s1 is None:
                    s1 = l_s1
                    p00 = l_p00
                else:
                    s1 = self.backend.bk_concat([s1, l_s1], axis=-1)
                    p00 = self.backend.bk_concat([p00, l_p00], axis=-1)

                # Concat l2_image [....,Npix_j1,....,j1]
                if l2_image is None:
                    l2_image = self.backend.bk_expand_dims(conj, axis=1)
                else:
                    l2_image = self.backend.bk_concat(
                        [self.backend.bk_expand_dims(conj, axis=1), l2_image], axis=1
                    )

            # Convol l2_image [....,Npix_j1,....,j1,Norient,Norient]
            c2_image = self.convol_1d(self.backend.bk_relu(l2_image), axis=axis + 2)

            conj2p = c2_image * self.backend.bk_conjugate(c2_image)
            conj2pl1 = self.backend.bk_L1(conj2p)

            if Auto:
                conj2p = self.backend.bk_real(conj2p)
                conj2pl1 = self.backend.bk_real(conj2pl1)

            c2_image = self.convol_1d(self.backend.bk_relu(-l2_image), axis=axis + 2)

            conj2m = c2_image * self.backend.bk_conjugate(c2_image)
            conj2ml1 = self.backend.bk_L1(conj2m)

            if Auto:
                conj2m = self.backend.bk_real(conj2m)
                conj2ml1 = self.backend.bk_real(conj2ml1)

            # Convol l_s2 [....,....,Nmask,j1]
            l_s2 = self.backend.bk_reduce_sum(
                (conj2p - conj2m) * self.backend.bk_expand_dims(vmask, 1), axis=axis + 2
            )
            l_s2l1 = self.backend.bk_reduce_sum(
                (conj2pl1 - conj2ml1) * self.backend.bk_expand_dims(vmask, 1),
                axis=axis + 2,
            )

            # Concat l_s2 [....,....,Nmask,j1*(j1+1)/2,Norient,Norient]
            if s2 is None:
                s2l = l_s2
                s2 = l_s2l1
                s2j1 = np.arange(l_s2.shape[axis], dtype="int")
                s2j2 = j1 * np.ones(l_s2.shape[axis], dtype="int")
            else:
                s2 = self.backend.bk_concat([s2, l_s2l1], axis=-1)
                s2l = self.backend.bk_concat([s2l, l_s2], axis=-1)
                s2j1 = np.concatenate(
                    [s2j1, np.arange(l_s2.shape[axis + 1], dtype="int")], 0
                )
                s2j2 = np.concatenate(
                    [s2j2, j1 * np.ones(l_s2.shape[axis + 1], dtype="int")], 0
                )

            if j1 != jmax - 1:
                # Rescale vmask [Nmask,Npix_j1//4]
                vmask = self.smooth_1d(vmask, axis=1)
                vmask = self.ud_grade_1d(vmask, vmask.shape[1] // 2, axis=1)
                if self.mask_thres is not None:
                    vmask = self.backend.bk_threshold(vmask, self.mask_thres)

                # Rescale l2_image [....,Npix_j1//4,....,j1,Norient]
                l2_image = self.smooth_1d(l2_image, axis=axis + 2)
                l2_image = self.ud_grade_1d(
                    l2_image, l2_image.shape[axis + 2] // 2, axis=axis + 2
                )

                # Rescale l_image [....,Npix_j1//4,....]
                l_image1 = self.smooth_1d(l_image1, axis=axis + 1)
                l_image1 = self.ud_grade_1d(
                    l_image1, l_image1.shape[axis + 1] // 2, axis=axis + 1
                )
                if cross:
                    l_image2 = self.smooth_1d(l_image2, axis=axis + 2)
                    l_image2 = self.ud_grade_1d(
                        l_image2, l_image2.shape[axis + 2] // 2, axis=axis + 2
                    )

        return scat1D(
            p00, s0, s1, s2, s2l, s2j1, s2j2, cross=cross, backend=self.backend
        )

    def square(self, x):
        # the abs make the complex value usable for reduce_sum or mean
        return scat1D(
            self.backend.bk_square(self.backend.bk_abs(x.P00)),
            self.backend.bk_square(self.backend.bk_abs(x.S0)),
            self.backend.bk_square(self.backend.bk_abs(x.S1)),
            self.backend.bk_square(self.backend.bk_abs(x.S2)),
            self.backend.bk_square(self.backend.bk_abs(x.S2L)),
            x.j1,
            x.j2,
            backend=self.backend,
        )

    def sqrt(self, x):
        # the abs make the complex value usable for reduce_sum or mean
        return scat1D(
            self.backend.bk_sqrt(self.backend.bk_abs(x.P00)),
            self.backend.bk_sqrt(self.backend.bk_abs(x.S0)),
            self.backend.bk_sqrt(self.backend.bk_abs(x.S1)),
            self.backend.bk_sqrt(self.backend.bk_abs(x.S2)),
            self.backend.bk_sqrt(self.backend.bk_abs(x.S2L)),
            x.j1,
            x.j2,
            backend=self.backend,
        )

    def reduce_mean(self, x, axis=None):
        if axis is None:
            tmp = (
                self.backend.bk_abs(self.backend.bk_reduce_sum(x.P00))
                + self.backend.bk_abs(self.backend.bk_reduce_sum(x.S0))
                + self.backend.bk_abs(self.backend.bk_reduce_sum(x.S1))
                + self.backend.bk_abs(self.backend.bk_reduce_sum(x.S2))
                + self.backend.bk_abs(self.backend.bk_reduce_sum(x.S2L))
            )

            ntmp = (
                np.array(list(x.P00.shape)).prod()
                + np.array(list(x.S0.shape)).prod()
                + np.array(list(x.S1.shape)).prod()
                + np.array(list(x.S2.shape)).prod()
            )

            return tmp / ntmp
        else:
            tmp = (
                self.backend.bk_abs(self.backend.bk_reduce_sum(x.P00, axis=axis))
                + self.backend.bk_abs(self.backend.bk_reduce_sum(x.S0, axis=axis))
                + self.backend.bk_abs(self.backend.bk_reduce_sum(x.S1, axis=axis))
                + self.backend.bk_abs(self.backend.bk_reduce_sum(x.S2, axis=axis))
                + self.backend.bk_abs(self.backend.bk_reduce_sum(x.S2L, axis=axis))
            )

            ntmp = (
                np.array(list(x.P00.shape)).prod()
                + np.array(list(x.S0.shape)).prod()
                + np.array(list(x.S1.shape)).prod()
                + np.array(list(x.S2.shape)).prod()
                + np.array(list(x.S2L.shape)).prod()
            )

            return tmp / ntmp

    def reduce_sum(self, x, axis=None):
        if axis is None:
            return (
                self.backend.bk_reduce_sum(self.backend.bk_abs(x.P00))
                + self.backend.bk_reduce_sum(self.backend.bk_abs(x.S0))
                + self.backend.bk_reduce_sum(self.backend.bk_abs(x.S1))
                + self.backend.bk_reduce_sum(self.backend.bk_abs(x.S2))
                + self.backend.bk_reduce_sum(self.backend.bk_abs(x.S2L))
            )
        else:
            return scat1D(
                self.backend.bk_reduce_sum(x.P00, axis=axis),
                self.backend.bk_reduce_sum(x.S0, axis=axis),
                self.backend.bk_reduce_sum(x.S1, axis=axis),
                self.backend.bk_reduce_sum(x.S2, axis=axis),
                self.backend.bk_reduce_sum(x.S2L, axis=axis),
                x.j1,
                x.j2,
                backend=self.backend,
            )

    def ldiff(self, sig, x):
        return scat1D(
            x.domult(sig.P00, x.P00) * x.domult(sig.P00, x.P00),
            x.domult(sig.S0, x.S0) * x.domult(sig.S0, x.S0),
            x.domult(sig.S1, x.S1) * x.domult(sig.S1, x.S1),
            x.domult(sig.S2, x.S2) * x.domult(sig.S2, x.S2),
            x.domult(sig.S2L, x.S2L) * x.domult(sig.S2L, x.S2L),
            x.j1,
            x.j2,
            backend=self.backend,
        )

    def log(self, x):
        return scat1D(
            self.backend.bk_log(x.P00),
            self.backend.bk_log(x.S0),
            self.backend.bk_log(x.S1),
            self.backend.bk_log(x.S2),
            self.backend.bk_log(x.S2L),
            x.j1,
            x.j2,
            backend=self.backend,
        )

    def abs(self, x):
        return scat1D(
            self.backend.bk_abs(x.P00),
            self.backend.bk_abs(x.S0),
            self.backend.bk_abs(x.S1),
            self.backend.bk_abs(x.S2),
            self.backend.bk_abs(x.S2L),
            x.j1,
            x.j2,
            backend=self.backend,
        )

    def inv(self, x):
        return scat1D(
            1 / (x.P00),
            1 / (x.S0),
            1 / (x.S1),
            1 / (x.S2),
            1 / (x.S2L),
            x.j1,
            x.j2,
            backend=self.backend,
        )

    def one(self):
        return scat1D(1.0, 1.0, 1.0, 1.0, 1.0, [0], [0], backend=self.backend)

    def eval_comp_fast(self, image1, image2=None, mask=None, Auto=True, s0_off=1e-6):

        res = self.eval_fast(image1, image2=image2, mask=mask, Auto=Auto, s0_off=s0_off)
        return res.P00, res.S0, res.S1, res.S2, res.S2L, res.j1, res.j2

    @tf_function
    def eval_fast(self, image1, image2=None, mask=None, Auto=True, s0_off=1e-6):
        p0, s0, s1, s2, s2l, j1, j2 = self.eval_comp_fast(
            image1, image2=image2, mask=mask, Auto=Auto, s0_off=s0_off
        )
        return scat1D(p0, s0, s1, s2, s2l, j1, j2, backend=self.backend)



import sys

import numpy as np
import tensorflow as tf

import foscat.BkBase as BackendBase


class BkTensorflow(BackendBase.BackendBase):

    def __init__(self, *args, **kwargs):
        # Impose que use_2D=True pour la classe scat
        super().__init__(name="tensorflow", *args, **kwargs)

        # ===========================================================================
        # INIT

        self.backend = tf
        # tf.config.threading.set_inter_op_parallelism_threads(1)
        # tf.config.threading.set_intra_op_parallelism_threads(1)
        self.tf_function = tf.function

        self.float64 = self.backend.float64
        self.float32 = self.backend.float32
        self.int64 = self.backend.int64
        self.int32 = self.backend.int32
        self.complex64 = self.backend.complex128
        self.complex128 = self.backend.complex64

        dtype_map = {
            "float32": (self.backend.float32, self.backend.complex64),
            "float64": (self.backend.float64, self.backend.complex128),
        }

        if self.all_type in dtype_map:
            self.all_bk_type, self.all_cbk_type = dtype_map[self.all_type]
        else:
            raise ValueError(
                f"ERROR INIT foscat: {self.all_type} should be float32 or float64"
            )

        if self.mpi_rank == 0:
            if not self.silent:
                print(
                    "Num GPUs Available: ",
                    len(self.backend.config.experimental.list_physical_devices("GPU")),
                )
            sys.stdout.flush()

        self.backend.debugging.set_log_device_placement(False)
        self.backend.config.set_soft_device_placement(True)

        gpus = self.backend.config.experimental.list_physical_devices("GPU")

        gpuname = "CPU:0"
        self.gpulist = {}
        self.gpulist[0] = gpuname
        self.ngpu = 1

        if gpus:
            try:
                # Currently, memory growth needs to be the same across GPUs
                for gpu in gpus:
                    self.backend.config.experimental.set_memory_growth(gpu, True)
                logical_gpus = self.backend.config.experimental.list_logical_devices(
                    "GPU"
                )
                print(len(gpus), "Physical GPUs,", len(logical_gpus), "Logical GPUs")
                sys.stdout.flush()
                self.ngpu = len(logical_gpus)
                if self.ngpu > 0:
                    gpuname = logical_gpus[self.gpupos % self.ngpu].name
                    self.gpulist = {}
                    for i in range(self.ngpu):
                        self.gpulist[i] = logical_gpus[i].name
                else:
                    gpuname = "CPU:0"
                    self.gpulist = {}
                    self.gpulist[0] = gpuname
                    self.ngpu = 1

            except RuntimeError as e:
                # Memory growth must be set before GPUs have been initialized
                print(e)

    def tf_loc_function(self, func):
        return func

    # ---------------------------------------------−---------
    # --             BACKEND DEFINITION                    --
    # ---------------------------------------------−---------
    def bk_SparseTensor(self, indice, w, dense_shape=[]):
        return self.backend.SparseTensor(indice, w, dense_shape=dense_shape)

    def bk_stack(self, list, axis=0):
        return self.backend.stack(list, axis=axis)

    def bk_sparse_dense_matmul(self, smat, mat):
        return self.backend.sparse.sparse_dense_matmul(smat, mat)

    # for tensorflow wrapping only
    def periodic_pad(self, x, pad_height, pad_width):
        """
        Applies periodic ('wrap') padding to a 4D TensorFlow tensor (N, H, W, C).

        Args:
        x (tf.Tensor): Input tensor with shape (batch_size, height, width, channels).
            pad_height (tuple): Tuple (top, bottom) defining the vertical padding size.
            pad_width (tuple): Tuple (left, right) defining the horizontal padding size.

        Returns:
            tf.Tensor: Tensor with periodic padding applied.
        """
        # Vertical padding: take slices from bottom and top to wrap around
        top_pad = x[:, -pad_height:, :, :]  # Top padding from the bottom rows
        bottom_pad = x[:, :pad_height, :, :]  # Bottom padding from the top rows
        x_padded = self.backend.concat(
            [top_pad, x, bottom_pad], axis=1
        )  # Concatenate vertically

        # Horizontal padding: take slices from right and left to wrap around
        left_pad = x_padded[:, :, -pad_width:, :]  # Left padding from right columns
        right_pad = x_padded[:, :, :pad_width, :]  # Right padding from left columns

        x_padded = self.backend.concat(
            [left_pad, x_padded, right_pad], axis=2
        )  # Concatenate horizontally

        return x_padded

    def binned_mean(self, data, cell_ids):
        """
        data: Tensor of shape [..., N] (float32 or float64)
        cell_ids: Tensor of shape [N], int indices in [0, n_bins)
        Returns: mean per bin, shape [..., n_bins]
        """
        ishape = list(data.shape)
        A = 1
        for k in range(len(ishape) - 1):
            A *= ishape[k]
        N = tf.shape(data)[-1]

        # Step 1: group indices
        groups = tf.math.floordiv(cell_ids, 4)  # [N]
        unique_groups, I = tf.unique(groups)  # I: [N]
        n_bins = tf.shape(unique_groups)[0]

        # Step 2: build I_tiled with batch + channel offsets
        I_tiled = tf.tile(I[None, :], [A, 1])  # shape [, N]

        # Offset index to flatten across [A, n_bins]
        batch_channel_offsets = tf.range(A)[:, None] * n_bins
        I_offset = I_tiled + batch_channel_offsets  # shape [A, N]]

        # Step 3: flatten data to shape [A, N]
        data_reshaped = tf.reshape(data, [A, N])  # shape [A, N]

        # Flatten all for scatter_nd
        indices = tf.reshape(I_offset, [-1])  # [A*N]
        values = tf.reshape(data_reshaped, [-1])  # [A*N]

        """
        # Prepare for scatter: indices → [A*N, 1]
        scatter_indices = tf.expand_dims(indices, axis=1)
        scatter_indices = tf.cast(scatter_indices, tf.int64)
        """
        total_bins = A * n_bins

        # Step 4: sum per bin
        sum_per_bin = tf.math.unsorted_segment_sum(values, indices, total_bins)
        sum_per_bin = tf.reshape(sum_per_bin, ishape[0:-1] + [n_bins])  # [A, n_bins]

        # Step 5: count per bin (same indices)
        counts = tf.math.unsorted_segment_sum(1.0 + 0 * values, indices, total_bins)
        # counts = tf.math.bincount(indices, minlength=total_bins, maxlength=total_bins)
        counts = tf.reshape(counts, ishape[0:-1] + [n_bins])
        # counts = tf.maximum(counts, 1)  # Avoid division by zero
        # counts = tf.cast(counts, dtype=data.dtype)

        # Step 6: mean
        mean_per_bin = sum_per_bin / counts  # [B, A, n_bins]

        return mean_per_bin, unique_groups

    def conv2d(self, x, w):
        """
        Perform 2D convolution using TensorFlow.

        Args:
            x: Tensor of shape [..., Nx, Ny] – input
            w: Tensor of shape [O_c, wx, wy] – conv weights

        Returns:
            Tensor of shape [..., O_c, Nx, Ny]
        """
        # Extract shape
        *leading_dims, Nx, Ny = x.shape
        O_c, wx, wy = w.shape

        # Flatten leading dims into a batch dimension
        B = tf.reduce_prod(leading_dims) if leading_dims else 1
        x = tf.reshape(x, [B, Nx, Ny, 1])  # TensorFlow format: [B, H, W, C_in=1]

        # Reshape weights to [wx, wy, in_channels=1, out_channels]
        w = tf.reshape(w, [O_c, wx, wy])
        w = tf.transpose(w, perm=[1, 2, 0])  # [wx, wy, O_c]
        w = tf.reshape(w, [wx, wy, 1, O_c])  # [wx, wy, C_in=1, C_out]

        # Apply 'reflect' padding manually
        pad_x = wx // 2
        pad_y = wy // 2
        x_padded = tf.pad(
            x, [[0, 0], [pad_x, pad_x], [pad_y, pad_y], [0, 0]], mode="REFLECT"
        )

        # Perform convolution
        y = tf.nn.conv2d(
            x_padded, w, strides=[1, 1, 1, 1], padding="VALID"
        )  # [B, Nx, Ny, O_c]

        # Transpose back to match original format: [..., O_c, Nx, Ny]
        y = tf.transpose(y, [0, 3, 1, 2])  # [B, O_c, Nx, Ny]
        y = tf.reshape(y, [*leading_dims, O_c, Nx, Ny])

        return y

    def conv1d(self, x, w):
        """
        Perform 1D convolution using TensorFlow.

        Args:
            x: Tensor of shape [..., N] – input
            w: Tensor of shape [k] – conv weights

        Returns:
            Tensor of shape [...,N]
        """
        # Extract shapes
        *leading_dims, N = x.shape
        k = w.shape[0]

        # Flatten leading dims into batch dimension
        B = tf.reduce_prod(leading_dims) if leading_dims else 1
        x = tf.reshape(x, [B, N, 1])  # TensorFlow 1D format: [B, L, C=1]

        # Prepare weights: [k, in_channels=1, out_channels=O_c]
        w = tf.reshape(w, [k, 1, 1])

        # Apply 'reflect' padding
        pad = k // 2
        x_padded = tf.pad(x, [[0, 0], [pad, pad], [0, 0]], mode="REFLECT")

        # Perform convolution
        y = tf.nn.conv1d(x_padded, w, stride=1, padding="VALID")  # [B, N, O_c]

        # Transpose to [B, O_c, N] and reshape back
        y = tf.transpose(y, [0, 2, 1])  # [B, 1, N]
        y = tf.reshape(y, [*leading_dims, N])  # [..., N]

        return y

    def bk_threshold(self, x, threshold, greater=True):

        return self.backend.cast(x > threshold, x.dtype) * x

    def bk_maximum(self, x1, x2):
        return self.backend.maximum(x1, x2)

    def bk_device(self, device_name):
        return self.backend.device(device_name)

    def bk_ones(self, shape, dtype=None):
        if dtype is None:
            dtype = self.all_type
        return self.backend.ones(shape, dtype=dtype)

    def bk_conv1d(self, x, w):
        return self.backend.nn.conv1d(x, w, stride=[1, 1, 1], padding="SAME")

    def bk_flattenR(self, x):
        if self.bk_is_complex(x):
            rr = self.backend.reshape(
                self.bk_real(x), [np.prod(np.array(list(x.shape)))]
            )
            ii = self.backend.reshape(
                self.bk_imag(x), [np.prod(np.array(list(x.shape)))]
            )
            return self.bk_concat([rr, ii], axis=0)
        else:
            return self.backend.reshape(x, [np.prod(np.array(list(x.shape)))])

    def bk_flatten(self, x):
        return self.backend.flatten(x)

    def bk_resize_image(self, x, shape):
        return self.bk_cast(self.backend.image.resize(x, shape, method="bilinear"))

    def bk_L1(self, x):
        if x.dtype == self.all_cbk_type:
            xr = self.bk_real(x)
            xi = self.bk_imag(x)

            r = self.backend.sign(xr) * self.backend.sqrt(self.backend.sign(xr) * xr)
            # return r
            i = self.backend.sign(xi) * self.backend.sqrt(self.backend.sign(xi) * xi)

            return self.bk_complex(r, i)
        else:
            return self.backend.sign(x) * self.backend.sqrt(self.backend.sign(x) * x)

    def bk_square_comp(self, x):
        xr = self.bk_real(x)
        xi = self.bk_imag(x)

        r = xr * xr
        i = xi * xi
        return self.bk_complex(r, i)

    def bk_reduce_sum(self, data, axis=None):

        if axis is None:
            return self.backend.reduce_sum(data)
        else:
            return self.backend.reduce_sum(data, axis=axis)

    # ---------------------------------------------−---------
    # return a tensor size

    def bk_size(self, data):
        return self.backend.size(data)

    def bk_reduce_mean(self, data, axis=None):

        if axis is None:
            return self.backend.reduce_mean(data)
        else:
            return self.backend.reduce_mean(data, axis=axis)

    def bk_reduce_min(self, data, axis=None):

        if axis is None:
            return self.backend.reduce_min(data)
        else:
            return self.backend.reduce_min(data, axis=axis)

    def bk_random_seed(self, value):

        return self.backend.random.set_seed(value)

    def bk_random_uniform(self, shape):

        return self.backend.random.uniform(shape)

    def bk_reduce_std(self, data, axis=None):
        if axis is None:
            r = self.backend.math.reduce_std(data)
        else:
            r = self.backend.math.reduce_std(data, axis=axis)
        if self.bk_is_complex(data):
            return self.bk_complex(r, 0 * r)
        else:
            return r

    def bk_sqrt(self, data):

        return self.backend.sqrt(self.backend.abs(data))

    def bk_abs(self, data):
        return self.backend.abs(data)

    def bk_is_complex(self, data):

        if isinstance(data, np.ndarray):
            return data.dtype == "complex64" or data.dtype == "complex128"
        return data.dtype.is_complex

    def bk_distcomp(self, data):
        if self.bk_is_complex(data):
            res = self.bk_square(self.bk_real(data)) + self.bk_square(
                self.bk_imag(data)
            )
            return res
        else:
            return self.bk_square(data)

    def bk_norm(self, data):
        if self.bk_is_complex(data):
            res = self.bk_square(self.bk_real(data)) + self.bk_square(
                self.bk_imag(data)
            )
            return self.bk_sqrt(res)

        else:
            return self.bk_abs(data)

    def bk_square(self, data):

        return self.backend.square(data)

    def bk_log(self, data):
        return self.backend.math.log(data)

    def bk_matmul(self, a, b):
        return self.backend.matmul(a, b)

    def bk_tensor(self, data):
        return self.backend.constant(data)

    def bk_shape_tensor(self, shape):
        return self.backend.tensor(shape=shape)

    def bk_complex(self, real, imag):
        return self.backend.dtypes.complex(real, imag)

    def bk_exp(self, data):

        return self.backend.exp(data)

    def bk_min(self, data):

        return self.backend.reduce_min(data)

    def bk_argmin(self, data):

        return self.backend.argmin(data)

    def bk_tanh(self, data):

        return self.backend.math.tanh(data)

    def bk_max(self, data):

        return self.backend.reduce_max(data)

    def bk_argmax(self, data):

        return self.backend.argmax(data)

    def bk_reshape(self, data, shape):
        return self.backend.reshape(data, shape)

    def bk_repeat(self, data, nn, axis=0):
        return self.backend.repeat(data, nn, axis=axis)

    def bk_tile(self, data, nn, axis=0):
        order = [1 for k in data.shape]
        order[axis] = nn
        return self.backend.tile(data, self.backend.constant(order, tf.int32))

    def bk_roll(self, data, nn, axis=0):
        return self.backend.roll(data, nn, axis=axis)

    def bk_expand_dims(self, data, axis=0):
        return self.backend.expand_dims(data, axis=axis)

    def bk_transpose(self, data, thelist):
        return self.backend.transpose(data, thelist)

    def bk_concat(self, data, axis=None):

        if axis is None:
            if data[0].dtype == self.all_cbk_type:
                ndata = len(data)
                xr = self.backend.concat([self.bk_real(data[k]) for k in range(ndata)])
                xi = self.backend.concat([self.bk_imag(data[k]) for k in range(ndata)])
                return self.bk_complex(xr, xi)
            else:
                return self.backend.concat(data)
        else:
            if data[0].dtype == self.all_cbk_type:
                ndata = len(data)
                xr = self.backend.concat(
                    [self.bk_real(data[k]) for k in range(ndata)], axis=axis
                )
                xi = self.backend.concat(
                    [self.bk_imag(data[k]) for k in range(ndata)], axis=axis
                )
                return self.bk_complex(xr, xi)
            else:
                return self.backend.concat(data, axis=axis)

    def bk_zeros(self, shape, dtype=None):
        return self.backend.zeros(shape, dtype=dtype)

    def bk_gather(self, data, idx, axis=0):
        return self.backend.gather(data, idx, axis=axis)

    def bk_reverse(self, data, axis=0):
        return self.backend.reverse(data, axis=[axis])

    def bk_fft(self, data):
        return self.backend.signal.fft(data)

    def bk_fftn(self, data, dim=None):
        # Equivalent of torch.fft.fftn(x, dim=dims) in TensorFlow
        if len(dim) == 2:
            return self.backend.signal.fft2d(self.bk_complex(data, 0 * data))
        else:
            return self.backend.signal.fft1d(self.bk_complex(data, 0 * data))

    def bk_ifftn(self, data, dim=None, norm=None):
        if norm is not None:
            if len(dim) == 2:
                normalization = self.backend.sqrt(
                    self.backend.cast(
                        data.shape[dim[0]] * data.shape[dim[1]], self.all_cbk_type
                    )
                )
                return self.backend.signal.ifft2d(data) * normalization

            else:
                normalization = self.backend.sqrt(
                    self.backend.cast(data.shape[dim[0]], self.all_cbk_type)
                )
                return self.backend.signal.ifft1d(data) * normalization
        else:
            if len(dim) == 2:
                return self.backend.signal.ifft2d(data)
            else:
                return self.backend.signal.ifft1d(data)

    def bk_rfft(self, data):
        return self.backend.signal.rfft(data)

    def bk_irfft(self, data):
        return self.backend.signal.irfft(data)

    def bk_conjugate(self, data):

        return self.backend.math.conj(data)

    def bk_real(self, data):
        return self.backend.math.real(data)

    def bk_imag(self, data):
        return self.backend.math.imag(data)

    def bk_relu(self, x):
        if x.dtype == self.all_cbk_type:
            xr = self.backend.nn.relu(self.bk_real(x))
            xi = self.backend.nn.relu(self.bk_imag(x))
            return self.bk_complex(xr, xi)
        else:
            return self.backend.nn.relu(x)

    def bk_clip_by_value(self, x, xmin, xmax):
        if isinstance(x, np.ndarray):
            x = np.clip(x, xmin, xmax)
        return self.backend.clip_by_value(x, xmin, xmax)

    def bk_cast(self, x):
        if isinstance(x, np.float64):
            if self.all_bk_type == "float32":
                return np.float32(x)
            else:
                return x
        if isinstance(x, np.float32):
            if self.all_bk_type == "float64":
                return np.float64(x)
            else:
                return x
        if isinstance(x, np.complex128):
            if self.all_bk_type == "float32":
                return np.complex64(x)
            else:
                return x
        if isinstance(x, np.complex64):
            if self.all_bk_type == "float64":
                return np.complex128(x)
            else:
                return x

        if isinstance(x, np.int32) or isinstance(x, np.int64) or isinstance(x, int):
            if self.all_bk_type == "float64":
                return np.float64(x)
            else:
                return np.float32(x)

        if self.bk_is_complex(x):
            out_type = self.all_cbk_type
        else:
            out_type = self.all_bk_type

        return self.backend.cast(x, out_type)

    def bk_variable(self, x):
        return self.backend.Variable(x)

    def bk_assign(self, x, y):
        return x.assign(y)

    def bk_constant(self, x):
        return self.backend.constant(x)

    def bk_cos(self, x):
        return self.backend.cos(x)

    def bk_sin(self, x):
        return self.backend.sin(x)

    def bk_arctan2(self, c, s):
        return self.backend.arctan2(c, s)

    def bk_empty(self, list):
        return self.backend.constant(list)

    def to_numpy(self, x):
        if isinstance(x, np.ndarray):
            return x

        return x.numpy()



import sys

import numpy as np


class foscat_backend:

    def __init__(self, name, mpi_rank=0, all_type="float64", gpupos=0, silent=False):

        self.TENSORFLOW = 1
        self.TORCH = 2
        self.NUMPY = 3

        # table use to compute the iso orientation rotation
        self._iso_orient = {}
        self._iso_orient_T = {}
        self._iso_orient_C = {}
        self._iso_orient_C_T = {}
        self._fft_1_orient = {}
        self._fft_1_orient_C = {}
        self._fft_2_orient = {}
        self._fft_2_orient_C = {}
        self._fft_3_orient = {}
        self._fft_3_orient_C = {}

        self.BACKEND = name

        if name not in ["tensorflow", "torch", "numpy"]:
            print('Backend "%s" not yet implemented' % (name))
            print(" Choose inside the next 3 available backends :")
            print(" - tensorflow")
            print(" - torch")
            print(" - numpy (Impossible to do synthesis using numpy)")
            return None

        if self.BACKEND == "tensorflow":
            import tensorflow as tf

            self.backend = tf
            self.BACKEND = self.TENSORFLOW
            # tf.config.threading.set_inter_op_parallelism_threads(1)
            # tf.config.threading.set_intra_op_parallelism_threads(1)
            self.tf_function = tf.function

        if self.BACKEND == "torch":
            import torch

            self.torch_device = (
                torch.device("cuda")
                if torch.cuda.is_available()
                else torch.device("cpu")
            )

            self.BACKEND = self.TORCH
            self.backend = torch
            self.tf_function = self.tf_loc_function

        if self.BACKEND == "numpy":
            self.BACKEND = self.NUMPY
            self.backend = np
            import scipy as scipy

            self.scipy = scipy
            self.tf_function = self.tf_loc_function

        self.float64 = self.backend.float64
        self.float32 = self.backend.float32
        self.int64 = self.backend.int64
        self.int32 = self.backend.int32
        self.complex64 = self.backend.complex128
        self.complex128 = self.backend.complex64

        if all_type == "float32":
            self.all_bk_type = self.backend.float32
            self.all_cbk_type = self.backend.complex64
        else:
            if all_type == "float64":
                self.all_type = "float64"
                self.all_bk_type = self.backend.float64
                self.all_cbk_type = self.backend.complex128
            else:
                print("ERROR INIT FOCUS ", all_type, " should be float32 or float64")
                return None
        # ===========================================================================
        # INIT
        if mpi_rank == 0:
            if self.BACKEND == self.TENSORFLOW and not silent:
                print(
                    "Num GPUs Available: ",
                    len(self.backend.config.experimental.list_physical_devices("GPU")),
                )
            sys.stdout.flush()

        if self.BACKEND == self.TENSORFLOW:
            self.backend.debugging.set_log_device_placement(False)
            self.backend.config.set_soft_device_placement(True)

            gpus = self.backend.config.experimental.list_physical_devices("GPU")

        if self.BACKEND == self.TORCH:
            gpus = torch.cuda.is_available()

        if self.BACKEND == self.NUMPY:
            gpus = []
        gpuname = "CPU:0"
        self.gpulist = {}
        self.gpulist[0] = gpuname
        self.ngpu = 1

        if gpus:
            try:
                if self.BACKEND == self.TENSORFLOW:
                    # Currently, memory growth needs to be the same across GPUs
                    for gpu in gpus:
                        self.backend.config.experimental.set_memory_growth(gpu, True)
                    logical_gpus = (
                        self.backend.config.experimental.list_logical_devices("GPU")
                    )
                    print(
                        len(gpus), "Physical GPUs,", len(logical_gpus), "Logical GPUs"
                    )
                    sys.stdout.flush()
                    self.ngpu = len(logical_gpus)
                    gpuname = logical_gpus[gpupos % self.ngpu].name
                    self.gpulist = {}
                    for i in range(self.ngpu):
                        self.gpulist[i] = logical_gpus[i].name
                if self.BACKEND == self.TORCH:
                    self.ngpu = torch.cuda.device_count()
                    self.gpulist = {}
                    for k in range(self.ngpu):
                        self.gpulist[k] = torch.cuda.get_device_name(0)

            except RuntimeError as e:
                # Memory growth must be set before GPUs have been initialized
                print(e)

    def tf_loc_function(self, func):
        return func

    def calc_iso_orient(self, norient):
        tmp = np.zeros([norient * norient, norient])
        for i in range(norient):
            for j in range(norient):
                tmp[j * norient + (j + i) % norient, i] = 0.25

        self._iso_orient[norient] = self.constant(self.bk_cast(tmp))
        self._iso_orient_T[norient] = self.constant(self.bk_cast(4 * tmp.T))
        self._iso_orient_C[norient] = self.bk_complex(
            self._iso_orient[norient], 0 * self._iso_orient[norient]
        )
        self._iso_orient_C_T[norient] = self.bk_complex(
            self._iso_orient_T[norient], 0 * self._iso_orient_T[norient]
        )

    def calc_fft_orient(self, norient, nharm, imaginary):

        x = np.arange(norient) / norient * 2 * np.pi

        if imaginary:
            tmp = np.zeros([norient, 1 + nharm * 2])
            tmp[:, 0] = 1.0
            for k in range(nharm):
                tmp[:, k * 2 + 1] = np.cos(x * (k + 1))
                tmp[:, k * 2 + 2] = np.sin(x * (k + 1))

            self._fft_1_orient[(norient, nharm, imaginary)] = self.bk_cast(
                self.constant(tmp)
            )
            self._fft_1_orient_C[(norient, nharm, imaginary)] = self.bk_complex(
                self._fft_1_orient[(norient, nharm, imaginary)],
                0 * self._fft_1_orient[(norient, nharm, imaginary)],
            )
        else:
            tmp = np.zeros([norient, 1 + nharm])
            for k in range(nharm + 1):
                tmp[:, k] = np.cos(x * k)

            self._fft_1_orient[(norient, nharm, imaginary)] = self.bk_cast(
                self.constant(tmp)
            )
            self._fft_1_orient_C[(norient, nharm, imaginary)] = self.bk_complex(
                self._fft_1_orient[(norient, nharm, imaginary)],
                0 * self._fft_1_orient[(norient, nharm, imaginary)],
            )

        x = np.repeat(x, norient).reshape(norient, norient)

        if imaginary:
            tmp = np.zeros([norient, norient, (1 + nharm * 2), (1 + nharm * 2)])
            tmp[:, :, 0, 0] = 1.0
            for k in range(nharm):
                tmp[:, :, k * 2 + 1, 0] = np.cos(x * (k + 1))
                tmp[:, :, k * 2 + 2, 0] = np.sin(x * (k + 1))
                tmp[:, :, 0, k * 2 + 1] = np.cos((x.T) * (k + 1))
                tmp[:, :, 0, k * 2 + 2] = np.sin((x.T) * (k + 1))
                for l_orient in range(nharm):
                    tmp[:, :, k * 2 + 1, l_orient * 2 + 1] = np.cos(
                        x * (k + 1)
                    ) * np.cos((x.T) * (l_orient + 1))
                    tmp[:, :, k * 2 + 2, l_orient * 2 + 1] = np.sin(
                        x * (k + 1)
                    ) * np.cos((x.T) * (l_orient + 1))
                    tmp[:, :, k * 2 + 1, l_orient * 2 + 2] = np.cos(
                        x * (k + 1)
                    ) * np.sin((x.T) * (l_orient + 1))
                    tmp[:, :, k * 2 + 2, l_orient * 2 + 2] = np.sin(
                        x * (k + 1)
                    ) * np.sin((x.T) * (l_orient + 1))

            self._fft_2_orient[(norient, nharm, imaginary)] = self.bk_cast(
                self.constant(
                    tmp.reshape(norient * norient, (1 + 2 * nharm) * (1 + 2 * nharm))
                )
            )
            self._fft_2_orient_C[(norient, nharm, imaginary)] = self.bk_complex(
                self._fft_2_orient[(norient, nharm, imaginary)],
                0 * self._fft_2_orient[(norient, nharm, imaginary)],
            )
        else:
            tmp = np.zeros([norient, norient, (1 + nharm), (1 + nharm)])

            for k in range(nharm + 1):
                for l_orient in range(nharm + 1):
                    tmp[:, :, k, l_orient] = np.cos(x * k) * np.cos((x.T) * l_orient)

            self._fft_2_orient[(norient, nharm, imaginary)] = self.bk_cast(
                self.constant(tmp.reshape(norient * norient, (1 + nharm) * (1 + nharm)))
            )
            self._fft_2_orient_C[(norient, nharm, imaginary)] = self.bk_complex(
                self._fft_2_orient[(norient, nharm, imaginary)],
                0 * self._fft_2_orient[(norient, nharm, imaginary)],
            )

        x = np.arange(norient) / norient * 2 * np.pi
        xx = np.zeros([norient, norient, norient])
        yy = np.zeros([norient, norient, norient])
        zz = np.zeros([norient, norient, norient])
        for i in range(norient):
            for j in range(norient):
                xx[:, i, j] = x
                yy[i, :, j] = x
                zz[i, j, :] = x

        if imaginary:
            tmp = np.ones(
                [
                    norient,
                    norient,
                    norient,
                    (1 + nharm * 2),
                    (1 + nharm * 2),
                    (1 + nharm * 2),
                ]
            )

            for k in range(nharm):
                tmp[:, :, :, k * 2 + 1, 0, 0] = np.cos(xx * (k + 1))
                tmp[:, :, :, 0, k * 2 + 1, 0] = np.cos(yy * (k + 1))
                tmp[:, :, :, 0, 0, k * 2 + 1] = np.cos(zz * (k + 1))

                tmp[:, :, :, k * 2 + 2, 0, 0] = np.sin(xx * (k + 1))
                tmp[:, :, :, 0, k * 2 + 2, 0] = np.sin(yy * (k + 1))
                tmp[:, :, :, 0, 0, k * 2 + 2] = np.sin(zz * (k + 1))
                for l_orient in range(nharm):
                    tmp[:, :, :, k * 2 + 1, l_orient * 2 + 1, 0] = np.cos(
                        xx * (k + 1)
                    ) * np.cos(yy * (l_orient + 1))
                    tmp[:, :, :, k * 2 + 1, l_orient * 2 + 2, 0] = np.cos(
                        xx * (k + 1)
                    ) * np.sin(yy * (l_orient + 1))
                    tmp[:, :, :, k * 2 + 2, l_orient * 2 + 1, 0] = np.sin(
                        xx * (k + 1)
                    ) * np.cos(yy * (l_orient + 1))
                    tmp[:, :, :, k * 2 + 2, l_orient * 2 + 2, 0] = np.sin(
                        xx * (k + 1)
                    ) * np.sin(yy * (l_orient + 1))

                    tmp[:, :, :, k * 2 + 1, 0, l_orient * 2 + 1] = np.cos(
                        xx * (k + 1)
                    ) * np.cos(zz * (l_orient + 1))
                    tmp[:, :, :, k * 2 + 1, 0, l_orient * 2 + 2] = np.cos(
                        xx * (k + 1)
                    ) * np.sin(zz * (l_orient + 1))
                    tmp[:, :, :, k * 2 + 2, 0, l_orient * 2 + 1] = np.sin(
                        xx * (k + 1)
                    ) * np.cos(zz * (l_orient + 1))
                    tmp[:, :, :, k * 2 + 2, 0, l_orient * 2 + 2] = np.sin(
                        xx * (k + 1)
                    ) * np.sin(zz * (l_orient + 1))

                    tmp[:, :, :, 0, k * 2 + 1, l_orient * 2 + 1] = np.cos(
                        yy * (k + 1)
                    ) * np.cos(zz * (l_orient + 1))
                    tmp[:, :, :, 0, k * 2 + 1, l_orient * 2 + 2] = np.cos(
                        yy * (k + 1)
                    ) * np.sin(zz * (l_orient + 1))
                    tmp[:, :, :, 0, k * 2 + 2, l_orient * 2 + 1] = np.sin(
                        yy * (k + 1)
                    ) * np.cos(zz * (l_orient + 1))
                    tmp[:, :, :, 0, k * 2 + 2, l_orient * 2 + 2] = np.sin(
                        yy * (k + 1)
                    ) * np.sin(zz * (l_orient + 1))

                    for m in range(nharm):
                        tmp[:, :, :, k * 2 + 1, l_orient * 2 + 1, m * 2 + 1] = (
                            np.cos(xx * (k + 1))
                            * np.cos(yy * (l_orient + 1))
                            * np.cos(zz * (m + 1))
                        )
                        tmp[:, :, :, k * 2 + 1, l_orient * 2 + 1, m * 2 + 2] = (
                            np.cos(xx * (k + 1))
                            * np.cos(yy * (l_orient + 1))
                            * np.sin(zz * (m + 1))
                        )
                        tmp[:, :, :, k * 2 + 1, l_orient * 2 + 2, m * 2 + 1] = (
                            np.cos(xx * (k + 1))
                            * np.sin(yy * (l_orient + 1))
                            * np.cos(zz * (m + 1))
                        )
                        tmp[:, :, :, k * 2 + 1, l_orient * 2 + 2, m * 2 + 2] = (
                            np.cos(xx * (k + 1))
                            * np.sin(yy * (l_orient + 1))
                            * np.sin(zz * (m + 1))
                        )
                        tmp[:, :, :, k * 2 + 2, l_orient * 2 + 1, m * 2 + 1] = (
                            np.sin(xx * (k + 1))
                            * np.cos(yy * (l_orient + 1))
                            * np.cos(zz * (m + 1))
                        )
                        tmp[:, :, :, k * 2 + 2, l_orient * 2 + 1, m * 2 + 2] = (
                            np.sin(xx * (k + 1))
                            * np.cos(yy * (l_orient + 1))
                            * np.sin(zz * (m + 1))
                        )
                        tmp[:, :, :, k * 2 + 2, l_orient * 2 + 2, m * 2 + 1] = (
                            np.sin(xx * (k + 1))
                            * np.sin(yy * (l_orient + 1))
                            * np.cos(zz * (m + 1))
                        )
                        tmp[:, :, :, k * 2 + 2, l_orient * 2 + 2, m * 2 + 2] = (
                            np.sin(xx * (k + 1))
                            * np.sin(yy * (l_orient + 1))
                            * np.sin(zz * (m + 1))
                        )

            self._fft_3_orient[(norient, nharm, imaginary)] = self.bk_cast(
                self.constant(
                    tmp.reshape(
                        norient * norient * norient,
                        (1 + nharm * 2) * (1 + nharm * 2) * (1 + nharm * 2),
                    )
                )
            )
            self._fft_3_orient_C[(norient, nharm, imaginary)] = self.bk_complex(
                self._fft_3_orient[(norient, nharm, imaginary)],
                0 * self._fft_3_orient[(norient, nharm, imaginary)],
            )
        else:
            tmp = np.zeros(
                [norient, norient, norient, (1 + nharm), (1 + nharm), (1 + nharm)]
            )

            for k in range(nharm + 1):
                for l_orient in range(nharm + 1):
                    for m in range(nharm + 1):
                        tmp[:, :, :, k, l_orient, m] = (
                            np.cos(xx * k) * np.cos(yy * l_orient) * np.cos(zz * m)
                        )

            self._fft_3_orient[(norient, nharm, imaginary)] = self.bk_cast(
                self.constant(
                    tmp.reshape(
                        norient * norient * norient,
                        (1 + nharm) * (1 + nharm) * (1 + nharm),
                    )
                )
            )
            self._fft_3_orient_C[(norient, nharm, imaginary)] = self.bk_complex(
                self._fft_3_orient[(norient, nharm, imaginary)],
                0 * self._fft_3_orient[(norient, nharm, imaginary)],
            )

    # ---------------------------------------------−---------
    # --             BACKEND DEFINITION                    --
    # ---------------------------------------------−---------
    def bk_SparseTensor(self, indice, w, dense_shape=[]):
        if self.BACKEND == self.TENSORFLOW:
            return self.backend.SparseTensor(indice, w, dense_shape=dense_shape)
        if self.BACKEND == self.TORCH:
            return (
                self.backend.sparse_coo_tensor(indice.T, w, dense_shape)
                .to_sparse_csr()
                .to(self.torch_device)
            )
        if self.BACKEND == self.NUMPY:
            return self.scipy.sparse.coo_matrix(
                (w, (indice[:, 0], indice[:, 1])), shape=dense_shape
            )

    def bk_stack(self, list, axis=0):
        if self.BACKEND == self.TENSORFLOW:
            return self.backend.stack(list, axis=axis)
        if self.BACKEND == self.TORCH:
            return self.backend.stack(list, axis=axis).to(self.torch_device)
        if self.BACKEND == self.NUMPY:
            return self.backend.stack(list, axis=axis)

    def bk_sparse_dense_matmul(self, smat, mat):
        if self.BACKEND == self.TENSORFLOW:
            return self.backend.sparse.sparse_dense_matmul(smat, mat)
        if self.BACKEND == self.TORCH:
            return smat.matmul(mat)
        if self.BACKEND == self.NUMPY:
            return smat.dot(mat)

    # for tensorflow wrapping only
    def periodic_pad(self, x, pad_height, pad_width):
        """
        Applies periodic ('wrap') padding to a 4D TensorFlow tensor (N, H, W, C).

        Args:
        x (tf.Tensor): Input tensor with shape (batch_size, height, width, channels).
            pad_height (tuple): Tuple (top, bottom) defining the vertical padding size.
            pad_width (tuple): Tuple (left, right) defining the horizontal padding size.

        Returns:
            tf.Tensor: Tensor with periodic padding applied.
        """
        # Vertical padding: take slices from bottom and top to wrap around
        top_pad = x[:, -pad_height:, :, :]  # Top padding from the bottom rows
        bottom_pad = x[:, :pad_height, :, :]  # Bottom padding from the top rows
        x_padded = self.backend.concat(
            [top_pad, x, bottom_pad], axis=1
        )  # Concatenate vertically

        # Horizontal padding: take slices from right and left to wrap around
        left_pad = x_padded[:, :, -pad_width:, :]  # Left padding from right columns
        right_pad = x_padded[:, :, :pad_width, :]  # Right padding from left columns

        x_padded = self.backend.concat(
            [left_pad, x_padded, right_pad], axis=2
        )  # Concatenate horizontally

        return x_padded

    def conv2d(self, x, w, strides=[1, 1, 1, 1], padding="SAME"):
        if self.BACKEND == self.TENSORFLOW:
            kx = w.shape[0]
            ky = w.shape[1]
            x_padded = self.periodic_pad(x, kx // 2, ky // 2)
            return self.backend.nn.conv2d(x_padded, w, strides=strides, padding="VALID")

        if self.BACKEND == self.TORCH:
            import torch.nn.functional as F

            lx = x.permute(0, 3, 1, 2)
            wx = (
                self.backend.from_numpy(w).to(self.torch_device).permute(3, 2, 0, 1)
            )  # de (5, 5, 1, 4) à (4, 1, 5, 5)

            # Calculer le padding symétrique
            kx, ky = w.shape[0], w.shape[1]

            # Appliquer le padding
            x_padded = F.pad(lx, (ky // 2, ky // 2, kx // 2, kx // 2), mode="circular")

            # Appliquer la convolution
            return F.conv2d(x_padded, wx, stride=1, padding=0).permute(0, 2, 3, 1)

        if self.BACKEND == self.NUMPY:
            res = np.zeros(
                [x.shape[0], x.shape[1], x.shape[2], w.shape[3]], dtype=x.dtype
            )
            for k in range(w.shape[2]):
                for l_orient in range(w.shape[3]):
                    for j in range(res.shape[0]):
                        tmp = self.scipy.signal.convolve2d(
                            x[j, :, :, k],
                            w[:, :, k, l_orient],
                            mode="same",
                            boundary="symm",
                        )
                        res[j, :, :, l_orient] += tmp
                        del tmp
            return res

    def conv1d(self, x, w, strides=[1, 1, 1], padding="SAME"):
        if self.BACKEND == self.TENSORFLOW:
            kx = w.shape[0]
            paddings = self.backend.constant([[0, 0], [kx // 2, kx // 2], [0, 0]])
            tmp = self.backend.pad(x, paddings, "SYMMETRIC")

            return self.backend.nn.conv1d(tmp, w, stride=strides, padding="VALID")
        # to be written!!!
        if self.BACKEND == self.TORCH:
            return x
        if self.BACKEND == self.NUMPY:
            res = np.zeros([x.shape[0], x.shape[1], w.shape[2]], dtype=x.dtype)
            for k in range(w.shape[2]):
                for j in range(res.shape[0]):
                    tmp = self.scipy.signal.convolve1d(
                        x[j, :, k], w[:, k], mode="same", boundary="symm"
                    )
                    res[j, :, :] += tmp
                    del tmp
            return res

    def bk_threshold(self, x, threshold, greater=True):

        if self.BACKEND == self.TENSORFLOW:
            return self.backend.cast(x > threshold, x.dtype) * x
        if self.BACKEND == self.TORCH:
            x.to(x.dtype)
            return (x > threshold) * x
            # return(self.backend.cast(x>threshold,x.dtype)*x)
        if self.BACKEND == self.NUMPY:
            return (x > threshold) * x

    def bk_maximum(self, x1, x2):
        if self.BACKEND == self.TENSORFLOW:
            return self.backend.maximum(x1, x2)
        if self.BACKEND == self.TORCH:
            return self.backend.maximum(x1, x2)
        if self.BACKEND == self.NUMPY:
            return x1 * (x1 > x2) + x2 * (x2 > x1)

    def bk_device(self, device_name):
        return self.backend.device(device_name)

    def bk_ones(self, shape, dtype=None):
        if dtype is None:
            dtype = self.all_type
        if self.BACKEND == self.TORCH:
            return self.bk_cast(np.ones(shape))
        return self.backend.ones(shape, dtype=dtype)

    def bk_conv1d(self, x, w):
        if self.BACKEND == self.TENSORFLOW:
            return self.backend.nn.conv1d(x, w, stride=[1, 1, 1], padding="SAME")
        if self.BACKEND == self.TORCH:
            # Torch not yet done !!!
            return self.backend.nn.conv1d(x, w, stride=1, padding="SAME")
        if self.BACKEND == self.NUMPY:
            res = np.zeros([x.shape[0], x.shape[1], w.shape[1]], dtype=x.dtype)
            for k in range(w.shape[1]):
                for l_orient in range(w.shape[2]):
                    res[:, :, l_orient] += self.scipy.ndimage.convolve1d(
                        x[:, :, k], w[:, k, l_orient], axis=1, mode="constant", cval=0.0
                    )
            return res

    def bk_flattenR(self, x):
        if self.BACKEND == self.TENSORFLOW or self.BACKEND == self.TORCH:
            if self.bk_is_complex(x):
                rr = self.backend.reshape(
                    self.bk_real(x), [np.prod(np.array(list(x.shape)))]
                )
                ii = self.backend.reshape(
                    self.bk_imag(x), [np.prod(np.array(list(x.shape)))]
                )
                return self.bk_concat([rr, ii], axis=0)
            else:
                return self.backend.reshape(x, [np.prod(np.array(list(x.shape)))])

        if self.BACKEND == self.NUMPY:
            if self.bk_is_complex(x):
                return np.concatenate([x.real.flatten(), x.imag.flatten()], 0)
            else:
                return x.flatten()

    def bk_flatten(self, x):
        if self.BACKEND == self.TENSORFLOW:
            return self.backend.reshape(x, [np.prod(np.array(list(x.shape)))])
        if self.BACKEND == self.TORCH:
            return self.backend.reshape(x, [np.prod(np.array(list(x.shape)))])
        if self.BACKEND == self.NUMPY:
            return x.flatten()

    def bk_resize_image(self, x, shape):
        if self.BACKEND == self.TENSORFLOW:
            return self.bk_cast(self.backend.image.resize(x, shape, method="bilinear"))

        if self.BACKEND == self.TORCH:
            tmp = self.backend.nn.functional.interpolate(
                x.permute(0, 3, 1, 2), size=shape, mode="bilinear", align_corners=False
            )
            return self.bk_cast(tmp.permute(0, 2, 3, 1))

        if self.BACKEND == self.NUMPY:
            return self.bk_cast(self.backend.image.resize(x, shape, method="bilinear"))

    def bk_L1(self, x):
        if x.dtype == self.all_cbk_type:
            xr = self.bk_real(x)
            xi = self.bk_imag(x)

            r = self.backend.sign(xr) * self.backend.sqrt(self.backend.sign(xr) * xr)
            # return r
            i = self.backend.sign(xi) * self.backend.sqrt(self.backend.sign(xi) * xi)

            if self.BACKEND == self.TORCH:
                return r
            else:
                return self.bk_complex(r, i)
        else:
            return self.backend.sign(x) * self.backend.sqrt(self.backend.sign(x) * x)

    def bk_square_comp(self, x):
        if x.dtype == self.all_cbk_type:
            xr = self.bk_real(x)
            xi = self.bk_imag(x)

            r = xr * xr
            i = xi * xi
            return self.bk_complex(r, i)
        else:
            return x * x

    def bk_reduce_sum(self, data, axis=None):

        if axis is None:
            if self.BACKEND == self.TENSORFLOW:
                return self.backend.reduce_sum(data)
            if self.BACKEND == self.TORCH:
                return self.backend.sum(data)
            if self.BACKEND == self.NUMPY:
                return np.sum(data)
        else:
            if self.BACKEND == self.TENSORFLOW:
                return self.backend.reduce_sum(data, axis=axis)
            if self.BACKEND == self.TORCH:
                return self.backend.sum(data, axis)
            if self.BACKEND == self.NUMPY:
                return np.sum(data, axis)

    # ---------------------------------------------−---------
    # return a tensor size

    def bk_size(self, data):
        if self.BACKEND == self.TENSORFLOW:
            return self.backend.size(data)
        if self.BACKEND == self.TORCH:
            return data.numel()
        if self.BACKEND == self.NUMPY:
            return data.size

    # ---------------------------------------------−---------

    def iso_mean(self, x, use_2D=False):
        shape = list(x.shape)

        i_orient = 2
        if use_2D:
            i_orient = 3
        norient = shape[i_orient]

        if len(shape) == i_orient + 1:
            return self.bk_reduce_mean(x, -1)

        if norient not in self._iso_orient:
            self.calc_iso_orient(norient)

        if self.bk_is_complex(x):
            lmat = self._iso_orient_C[norient]
        else:
            lmat = self._iso_orient[norient]

        oshape = shape[0]
        for k in range(1, len(shape) - 2):
            oshape *= shape[k]

        oshape2 = [shape[k] for k in range(0, len(shape) - 1)]

        return self.bk_reshape(
            self.backend.matmul(self.bk_reshape(x, [oshape, norient * norient]), lmat),
            oshape2,
        )

    def fft_ang(self, x, nharm=1, imaginary=False, use_2D=False):
        shape = list(x.shape)

        i_orient = 2
        if use_2D:
            i_orient = 3

        norient = shape[i_orient]
        nout = 1 + nharm

        oshape_1 = shape[0]
        for k in range(1, i_orient):
            oshape_1 *= shape[k]
        oshape_2 = norient
        for k in range(i_orient, len(shape) - 1):
            oshape_2 *= shape[k]
        oshape = [oshape_1, oshape_2]

        if imaginary:
            nout = 1 + nharm * 2

        oshape2 = [shape[k] for k in range(0, i_orient)] + [
            nout for k in range(i_orient, len(shape))
        ]

        if (norient, nharm) not in self._fft_1_orient:
            self.calc_fft_orient(norient, nharm, imaginary)

        if len(shape) == i_orient + 1:
            if self.bk_is_complex(x):
                lmat = self._fft_1_orient_C[(norient, nharm, imaginary)]
            else:
                lmat = self._fft_1_orient[(norient, nharm, imaginary)]

        if len(shape) == i_orient + 2:
            if self.bk_is_complex(x):
                lmat = self._fft_2_orient_C[(norient, nharm, imaginary)]
            else:
                lmat = self._fft_2_orient[(norient, nharm, imaginary)]

        if len(shape) == i_orient + 3:
            if self.bk_is_complex(x):
                lmat = self._fft_3_orient_C[(norient, nharm, imaginary)]
            else:
                lmat = self._fft_3_orient[(norient, nharm, imaginary)]

        return self.bk_reshape(
            self.backend.matmul(self.bk_reshape(x, oshape), lmat), oshape2
        )

    def constant(self, data):

        if self.BACKEND == self.TENSORFLOW:
            return self.backend.constant(data)
        return data

    def bk_reduce_mean(self, data, axis=None):

        if axis is None:
            if self.BACKEND == self.TENSORFLOW:
                return self.backend.reduce_mean(data)
            if self.BACKEND == self.TORCH:
                return self.backend.mean(data)
            if self.BACKEND == self.NUMPY:
                return np.mean(data)
        else:
            if self.BACKEND == self.TENSORFLOW:
                return self.backend.reduce_mean(data, axis=axis)
            if self.BACKEND == self.TORCH:
                return self.backend.mean(data, axis)
            if self.BACKEND == self.NUMPY:
                return np.mean(data, axis)

    def bk_reduce_min(self, data, axis=None):

        if axis is None:
            if self.BACKEND == self.TENSORFLOW:
                return self.backend.reduce_min(data)
            if self.BACKEND == self.TORCH:
                return self.backend.min(data)
            if self.BACKEND == self.NUMPY:
                return np.min(data)
        else:
            if self.BACKEND == self.TENSORFLOW:
                return self.backend.reduce_min(data, axis=axis)
            if self.BACKEND == self.TORCH:
                return self.backend.min(data, axis)
            if self.BACKEND == self.NUMPY:
                return np.min(data, axis)

    def bk_random_seed(self, value):

        if self.BACKEND == self.TENSORFLOW:
            return self.backend.random.set_seed(value)
        if self.BACKEND == self.TORCH:
            return self.backend.random.set_seed(value)
        if self.BACKEND == self.NUMPY:
            return np.random.seed(value)

    def bk_random_uniform(self, shape):

        if self.BACKEND == self.TENSORFLOW:
            return self.backend.random.uniform(shape)
        if self.BACKEND == self.TORCH:
            return self.backend.random.uniform(shape)
        if self.BACKEND == self.NUMPY:
            return np.random.rand(shape)

    def bk_reduce_std(self, data, axis=None):
        if axis is None:
            if self.BACKEND == self.TENSORFLOW:
                r = self.backend.math.reduce_std(data)
            if self.BACKEND == self.TORCH:
                r = self.backend.std(data)
            if self.BACKEND == self.NUMPY:
                r = np.std(data)
            return self.bk_complex(r, 0 * r)
        else:
            if self.BACKEND == self.TENSORFLOW:
                r = self.backend.math.reduce_std(data, axis=axis)
            if self.BACKEND == self.TORCH:
                r = self.backend.std(data, axis)
            if self.BACKEND == self.NUMPY:
                r = np.std(data, axis)
        if self.bk_is_complex(data):
            return self.bk_complex(r, 0 * r)
        else:
            return r

    def bk_sqrt(self, data):

        return self.backend.sqrt(self.backend.abs(data))

    def bk_abs(self, data):
        return self.backend.abs(data)

    def bk_is_complex(self, data):

        if self.BACKEND == self.TENSORFLOW:
            if isinstance(data, np.ndarray):
                return data.dtype == "complex64" or data.dtype == "complex128"
            return data.dtype.is_complex

        if self.BACKEND == self.TORCH:
            if isinstance(data, np.ndarray):
                return data.dtype == "complex64" or data.dtype == "complex128"

            return data.dtype.is_complex

        if self.BACKEND == self.NUMPY:
            return data.dtype == "complex64" or data.dtype == "complex128"

    def bk_distcomp(self, data):
        if self.bk_is_complex(data):
            res = self.bk_square(self.bk_real(data)) + self.bk_square(
                self.bk_imag(data)
            )
            return res
        else:
            return self.bk_square(data)

    def bk_norm(self, data):
        if self.bk_is_complex(data):
            res = self.bk_square(self.bk_real(data)) + self.bk_square(
                self.bk_imag(data)
            )
            return self.bk_sqrt(res)

        else:
            return self.bk_abs(data)

    def bk_square(self, data):

        if self.BACKEND == self.TENSORFLOW:
            return self.backend.square(data)
        if self.BACKEND == self.TORCH:
            return self.backend.square(data)
        if self.BACKEND == self.NUMPY:
            return data * data

    def bk_log(self, data):
        if self.BACKEND == self.TENSORFLOW:
            return self.backend.math.log(data)
        if self.BACKEND == self.TORCH:
            return self.backend.log(data)
        if self.BACKEND == self.NUMPY:
            return np.log(data)

    def bk_matmul(self, a, b):
        if self.BACKEND == self.TENSORFLOW:
            return self.backend.matmul(a, b)
        if self.BACKEND == self.TORCH:
            return self.backend.matmul(a, b)
        if self.BACKEND == self.NUMPY:
            return np.dot(a, b)

    def bk_tensor(self, data):
        if self.BACKEND == self.TENSORFLOW:
            return self.backend.constant(data)
        if self.BACKEND == self.TORCH:
            return self.backend.constant(data).to(self.torch_device)
        if self.BACKEND == self.NUMPY:
            return data

    def bk_shape_tensor(self, shape):
        if self.BACKEND == self.TENSORFLOW:
            return self.backend.tensor(shape=shape)
        if self.BACKEND == self.TORCH:
            return self.backend.tensor(shape=shape).to(self.torch_device)
        if self.BACKEND == self.NUMPY:
            return np.zeros(shape)

    def bk_complex(self, real, imag):
        if self.BACKEND == self.TENSORFLOW:
            return self.backend.dtypes.complex(real, imag)
        if self.BACKEND == self.TORCH:
            return self.backend.complex(real, imag).to(self.torch_device)
        if self.BACKEND == self.NUMPY:
            return real + 1j * imag

    def bk_exp(self, data):

        return self.backend.exp(data)

    def bk_min(self, data):

        return self.backend.reduce_min(data)

    def bk_argmin(self, data):

        return self.backend.argmin(data)

    def bk_tanh(self, data):

        return self.backend.math.tanh(data)

    def bk_max(self, data):

        return self.backend.reduce_max(data)

    def bk_argmax(self, data):

        return self.backend.argmax(data)

    def bk_reshape(self, data, shape):
        if self.BACKEND == self.TORCH:
            if isinstance(data, np.ndarray):
                return data.reshape(shape)
            return data.view(shape)

        return self.backend.reshape(data, shape)

    def bk_repeat(self, data, nn, axis=0):
        return self.backend.repeat(data, nn, axis=axis)

    def bk_tile(self, data, nn, axis=0):
        if self.BACKEND == self.TENSORFLOW:
            return self.backend.tile(data, [nn])

        return self.backend.tile(data, nn)

    def bk_roll(self, data, nn, axis=0):
        return self.backend.roll(data, nn, axis=axis)

    def bk_expand_dims(self, data, axis=0):
        if self.BACKEND == self.TENSORFLOW:
            return self.backend.expand_dims(data, axis=axis)
        if self.BACKEND == self.TORCH:
            if isinstance(data, np.ndarray):
                data = self.backend.from_numpy(data)
            return self.backend.unsqueeze(data, axis)
        if self.BACKEND == self.NUMPY:
            return np.expand_dims(data, axis)

    def bk_transpose(self, data, thelist):
        if self.BACKEND == self.TENSORFLOW:
            return self.backend.transpose(data, thelist)
        if self.BACKEND == self.TORCH:
            return self.backend.transpose(data, thelist)
        if self.BACKEND == self.NUMPY:
            return np.transpose(data, thelist)

    def bk_concat(self, data, axis=None):

        if self.BACKEND == self.TENSORFLOW or self.BACKEND == self.TORCH:
            if axis is None:
                if data[0].dtype == self.all_cbk_type:
                    ndata = len(data)
                    xr = self.backend.concat(
                        [self.bk_real(data[k]) for k in range(ndata)]
                    )
                    xi = self.backend.concat(
                        [self.bk_imag(data[k]) for k in range(ndata)]
                    )
                    return self.bk_complex(xr, xi)
                else:
                    return self.backend.concat(data)
            else:
                if data[0].dtype == self.all_cbk_type:
                    ndata = len(data)
                    xr = self.backend.concat(
                        [self.bk_real(data[k]) for k in range(ndata)], axis=axis
                    )
                    xi = self.backend.concat(
                        [self.bk_imag(data[k]) for k in range(ndata)], axis=axis
                    )
                    return self.bk_complex(xr, xi)
                else:
                    return self.backend.concat(data, axis=axis)
        else:
            if axis is None:
                return np.concatenate(data, axis=0)
            else:
                return np.concatenate(data, axis=axis)

    def bk_zeros(self, shape, dtype=None):
        if self.BACKEND == self.TENSORFLOW:
            return self.backend.zeros(shape, dtype=dtype)
        if self.BACKEND == self.TORCH:
            return self.backend.zeros(shape, dtype=dtype).to(self.torch_device)
        if self.BACKEND == self.NUMPY:
            return np.zeros(shape, dtype=dtype)

    def bk_gather(self, data, idx):
        if self.BACKEND == self.TENSORFLOW:
            return self.backend.gather(data, idx)
        if self.BACKEND == self.TORCH:
            return data[idx]
        if self.BACKEND == self.NUMPY:
            return data[idx]

    def bk_reverse(self, data, axis=0):
        if self.BACKEND == self.TENSORFLOW:
            return self.backend.reverse(data, axis=[axis])
        if self.BACKEND == self.TORCH:
            return self.backend.reverse(data, axis=axis)
        if self.BACKEND == self.NUMPY:
            return np.reverse(data, axis=axis)

    def bk_fft(self, data):
        if self.BACKEND == self.TENSORFLOW:
            return self.backend.signal.fft(data)
        if self.BACKEND == self.TORCH:
            return self.backend.fft(data)
        if self.BACKEND == self.NUMPY:
            return self.backend.fft.fft(data)

    def bk_fftn(self, data, dim=None):
        if self.BACKEND == self.TENSORFLOW:
            # Equivalent of torch.fft.fftn(x, dim=dims) in TensorFlow
            x = self.bk_complex(data, 0 * data)
            return self.backend.signal.fftnd(
                x, fft_length=tuple(x.shape[d] for d in dim), axes=dim
            )
        if self.BACKEND == self.TORCH:
            return self.backend.fft.fftn(data, dim=dim)
        if self.BACKEND == self.NUMPY:
            return self.backend.fft.fftn(data)

    def bk_ifftn(self, data, dim=None, norm=None):
        if self.BACKEND == self.TENSORFLOW:
            return self.backend.signal.ifftnd(
                data, fft_length=tuple(data.shape[d] for d in dim), axes=dim, norm=norm
            )
        if self.BACKEND == self.TORCH:
            return self.backend.fft.ifftn(data, dim=dim, norm=norm)
        if self.BACKEND == self.NUMPY:
            return self.backend.fft.ifftn(data)

    def bk_rfft(self, data):
        if self.BACKEND == self.TENSORFLOW:
            return self.backend.signal.rfft(data)
        if self.BACKEND == self.TORCH:
            return self.backend.rfft(data)
        if self.BACKEND == self.NUMPY:
            return self.backend.fft.rfft(data)

    def bk_irfft(self, data):
        if self.BACKEND == self.TENSORFLOW:
            return self.backend.signal.irfft(data)
        if self.BACKEND == self.TORCH:
            return self.backend.irfft(data)
        if self.BACKEND == self.NUMPY:
            return self.backend.fft.irfft(data)

    def bk_conjugate(self, data):

        if self.BACKEND == self.TENSORFLOW:
            return self.backend.math.conj(data)
        if self.BACKEND == self.TORCH:
            return self.backend.conj(data)
        if self.BACKEND == self.NUMPY:
            return data.conjugate()

    def bk_real(self, data):
        if self.BACKEND == self.TENSORFLOW:
            return self.backend.math.real(data)
        if self.BACKEND == self.TORCH:
            return data.real
        if self.BACKEND == self.NUMPY:
            return data.real

    def bk_imag(self, data):
        if self.BACKEND == self.TENSORFLOW:
            return self.backend.math.imag(data)
        if self.BACKEND == self.TORCH:
            if data.dtype == self.all_cbk_type:
                return data.imag
            else:
                return 0

        if self.BACKEND == self.NUMPY:
            return data.imag

    def bk_relu(self, x):
        if self.BACKEND == self.TENSORFLOW:
            if x.dtype == self.all_cbk_type:
                xr = self.backend.nn.relu(self.bk_real(x))
                xi = self.backend.nn.relu(self.bk_imag(x))
                return self.bk_complex(xr, xi)
            else:
                return self.backend.nn.relu(x)
        if self.BACKEND == self.TORCH:
            return self.backend.relu(x)
        if self.BACKEND == self.NUMPY:
            return (x > 0) * x

    def bk_clip_by_value(self, x, xmin, xmax):
        if isinstance(x, np.ndarray):
            x = np.clip(x, xmin, xmax)
        if self.BACKEND == self.TENSORFLOW:
            return self.backend.clip_by_value(x, xmin, xmax)
        if self.BACKEND == self.TORCH:
            x = (
                self.backend.tensor(x, dtype=self.backend.float32)
                if not isinstance(x, self.backend.Tensor)
                else x
            )
            xmin = (
                self.backend.tensor(xmin, dtype=self.backend.float32)
                if not isinstance(xmin, self.backend.Tensor)
                else xmin
            )
            xmax = (
                self.backend.tensor(xmax, dtype=self.backend.float32)
                if not isinstance(xmax, self.backend.Tensor)
                else xmax
            )
            return self.backend.clamp(x, min=xmin, max=xmax)
        if self.BACKEND == self.NUMPY:
            return self.backend.clip(x, xmin, xmax)

    def bk_cast(self, x):
        if isinstance(x, np.float64):
            if self.all_bk_type == "float32":
                return np.float32(x)
            else:
                return x
        if isinstance(x, np.float32):
            if self.all_bk_type == "float64":
                return np.float64(x)
            else:
                return x
        if isinstance(x, np.complex128):
            if self.all_bk_type == "float32":
                return np.complex64(x)
            else:
                return x
        if isinstance(x, np.complex64):
            if self.all_bk_type == "float64":
                return np.complex128(x)
            else:
                return x

        if isinstance(x, np.int32) or isinstance(x, np.int64) or isinstance(x, int):
            if self.all_bk_type == "float64":
                return np.float64(x)
            else:
                return np.float32(x)

        if self.bk_is_complex(x):
            out_type = self.all_cbk_type
        else:
            out_type = self.all_bk_type

        if self.BACKEND == self.TENSORFLOW:
            return self.backend.cast(x, out_type)

        if self.BACKEND == self.TORCH:
            if isinstance(x, np.ndarray):
                x = self.backend.from_numpy(x).to(self.torch_device)

            if x.dtype.is_complex:
                out_type = self.all_cbk_type
            else:
                out_type = self.all_bk_type

            return x.type(out_type).to(self.torch_device)

        if self.BACKEND == self.NUMPY:
            return x.astype(out_type)

    def bk_variable(self, x):
        if self.BACKEND == self.TENSORFLOW:
            return self.backend.Variable(x)

        return self.bk_cast(x)

    def bk_assign(self, x, y):
        if self.BACKEND == self.TENSORFLOW:
            x.assign(y)
        x = y

    def bk_constant(self, x):
        if self.BACKEND == self.TENSORFLOW:
            return self.backend.constant(x)

        return self.bk_cast(x)

    def to_numpy(self, x):
        if isinstance(x, np.ndarray):
            return x

        if self.BACKEND == self.NUMPY:
            return x
        if self.BACKEND == self.TENSORFLOW:
            return x.numpy()

        if self.BACKEND == self.TORCH:
            return x.cpu().numpy()



import torch


class loss_backend:

    def __init__(self, backend, curr_gpu, mpi_rank):

        self.bk = backend
        self.curr_gpu = curr_gpu
        self.mpi_rank = mpi_rank

    def check_dense(self, data, datasz):
        if isinstance(data, torch.Tensor):
            return data
        """
        idx=tf.cast(data.indices, tf.int32)
        data=tf.math.bincount(idx,weights=data.values,
                              minlength=datasz)
        """
        return data

    # ---------------------------------------------−---------

    def loss(self, x, batch, loss_function, KEEP_TRACK):

        operation = loss_function.scat_operator

        if torch.cuda.is_available():
            with torch.cuda.device((operation.gpupos + self.curr_gpu) % operation.ngpu):

                l_x = x.clone().detach().requires_grad_(True)

                if KEEP_TRACK is not None:
                    l_loss, linfo = loss_function.eval(l_x, batch, return_all=True)
                else:
                    l_loss = loss_function.eval(l_x, batch)

                l_loss.backward()

                g = l_x.grad

                self.curr_gpu = self.curr_gpu + 1
        else:
            l_x = x.clone().detach().requires_grad_(True)

            if KEEP_TRACK is not None:
                l_loss, linfo = loss_function.eval(l_x, batch, return_all=True)
            else:
                l_loss = loss_function.eval(l_x, batch)

            l_loss.backward()

            g = l_x.grad

        if KEEP_TRACK is not None:
            return l_loss.detach(), g, linfo
        else:
            return l_loss.detach(), g
